import pytest

from edsnlp.utils.batching import (
    DATASET_END_SENTINEL,
    BatchSizeArg,
    FragmentEndSentinel,
    StreamSentinel,
    batchify,
    batchify_by_dataset,
    batchify_by_fragment,
    batchify_by_length_sum,
    batchify_by_padded,
    stat_batchify,
)


class MockStreamSentinel(StreamSentinel):
    pass


# Tests for BatchSizeArg
def test_batch_size_arg_validate():
    # Valid inputs
    assert BatchSizeArg.validate("10 samples") == (10, "samples")
    assert BatchSizeArg.validate("20 words") == (20, "words")
    assert BatchSizeArg.validate(15) == (15, "docs")
    assert BatchSizeArg.validate("docs") == (None, "docs")
    assert BatchSizeArg.validate("tokens") == (None, "tokens")
    assert BatchSizeArg.validate("25") == (25, "docs")

    # Invalid inputs
    with pytest.raises(Exception):
        BatchSizeArg.validate("invalid input")
    with pytest.raises(Exception):
        BatchSizeArg.validate("10 invalid input")
    with pytest.raises(Exception):
        BatchSizeArg.validate("invalid input 10")


# Tests for batchify function
def test_batchify_simple():
    data = [1, 2, 3, 4, 5]
    batches = list(batchify(data, batch_size=2))
    assert batches == [[1, 2], [3, 4], [5]]


def test_batchify_drop_last():
    data = [1, 2, 3, 4, 5]
    batches = list(batchify(data, batch_size=2, drop_last=True))
    assert batches == [[1, 2], [3, 4]]


def test_batchify_sentinel_drop():
    data = [1, 2, MockStreamSentinel(), 3, 4]
    batches = list(batchify(data, batch_size=2, sentinel_mode="drop"))
    assert batches == [[1, 2], [3, 4]]


def test_batchify_sentinel_keep():
    sentinel = MockStreamSentinel()
    data = [1, 2, sentinel, 3, 4]
    batches = list(batchify(data, batch_size=2, sentinel_mode="keep"))
    assert batches == [[1, 2, sentinel], [3, 4]]


def test_batchify_sentinel_split():
    sentinel = MockStreamSentinel()
    data = [1, 2, sentinel, 3, 4]
    batches = list(batchify(data, batch_size=2, sentinel_mode="split"))
    assert batches == [[1, 2], sentinel, [3, 4]]


# Tests for batchify_by_length_sum
def test_batchify_by_length_sum_simple():
    data = ["a", "bb", "ccc", "dddd", "eeeee"]
    batches = list(batchify_by_length_sum(data, batch_size=5))
    assert batches == [["a", "bb"], ["ccc"], ["dddd"], ["eeeee"]]


def test_batchify_by_length_sum_drop_last():
    data = ["a", "bb", "ccc", "dddd", "eeeee"]
    batches = list(batchify_by_length_sum(data, batch_size=5, drop_last=True))
    assert batches == [["a", "bb"], ["ccc"], ["dddd"]]


# Tests for batchify_by_length_sum
def test_batchify_by_length_sum_split():
    sentinel = MockStreamSentinel()
    data = ["aa", "bb", sentinel, "ccc", "dddd", "eeeee"]
    batches = list(batchify_by_length_sum(data, batch_size=7, sentinel_mode="split"))
    assert batches == [["aa", "bb"], sentinel, ["ccc", "dddd"], ["eeeee"]]


# Tests for batchify_by_length_sum
def test_batchify_by_length_sum_keep():
    sentinel = MockStreamSentinel()
    data = ["aa", "bb", sentinel, "ccc", "dddd", "eeeee"]
    batches = list(batchify_by_length_sum(data, batch_size=7, sentinel_mode="keep"))
    assert batches == [["aa", "bb", sentinel, "ccc"], ["dddd"], ["eeeee"]]


# Tests for batchify_by_padded
def test_batchify_by_padded_simple():
    data = ["a", "bb", "ccc", "dddd"]
    batches = list(batchify_by_padded(data, batch_size=6))
    assert batches == [["a", "bb"], ["ccc"], ["dddd"]]


def test_batchify_by_padded_drop_last():
    data = ["a", "bb", "ccc", "dddd"]
    batches = list(batchify_by_padded(data, batch_size=6, drop_last=True))
    assert batches == [["a", "bb"], ["ccc"]]


def test_batchify_by_padded_sentinel_keep():
    sentinel = MockStreamSentinel()
    data = ["a", sentinel, "bb", "ccc"]
    batches = list(batchify_by_padded(data, batch_size=6, sentinel_mode="keep"))
    assert batches == [["a", sentinel, "bb"], ["ccc"]]


def test_batchify_by_padded_sentinel_split():
    sentinel = MockStreamSentinel()
    data = ["a", sentinel, "bb", "ccc"]
    batches = list(batchify_by_padded(data, batch_size=5, sentinel_mode="split"))
    assert batches == [["a"], sentinel, ["bb"], ["ccc"]]


# Tests for batchify_by_dataset
def test_batchify_by_dataset_simple():
    data = [
        "item1",
        "item2",
        DATASET_END_SENTINEL,
        "item3",
        DATASET_END_SENTINEL,
        "item4",
        "item5",
    ]
    batches = list(batchify_by_dataset(data))
    assert batches == [
        ["item1", "item2"],
        DATASET_END_SENTINEL,
        ["item3"],
        DATASET_END_SENTINEL,
        ["item4", "item5"],
    ]


def test_batchify_by_dataset_sentinel_split():
    sentinel = MockStreamSentinel()
    data = ["item1", sentinel, "item2", DATASET_END_SENTINEL, "item3"]
    batches = list(batchify_by_dataset(data, sentinel_mode="split"))
    assert batches == [["item1"], sentinel, ["item2"], DATASET_END_SENTINEL, ["item3"]]


def test_batchify_by_dataset_sentinel_keep():
    sentinel = MockStreamSentinel()
    data = ["item1", sentinel, "item2", DATASET_END_SENTINEL, "item3"]
    batches = list(batchify_by_dataset(data, sentinel_mode="keep"))
    assert batches == [["item1", sentinel, "item2"], DATASET_END_SENTINEL, ["item3"]]


def test_batchify_by_dataset_sentinel_drop():
    sentinel = MockStreamSentinel()
    data = ["item1", sentinel, "item2", DATASET_END_SENTINEL, "item3"]
    batches = list(batchify_by_dataset(data, sentinel_mode="drop"))
    assert batches == [["item1", "item2"], DATASET_END_SENTINEL, ["item3"]]


def test_batchify_by_dataset_drop_last():
    data = ["item1", "item2", DATASET_END_SENTINEL, "item3"]
    batches = list(batchify_by_dataset(data, drop_last=True))
    assert batches == [["item1", "item2"], DATASET_END_SENTINEL]


# Tests for batchify_by_fragment
def test_batchify_by_fragment_simple():
    fragment_end_1 = FragmentEndSentinel("fragment1")
    fragment_end_2 = FragmentEndSentinel("fragment2")
    data = ["item1", "item2", fragment_end_1, "item3", fragment_end_2, "item4"]
    batches = list(batchify_by_fragment(data))
    assert batches == [
        ["item1", "item2"],
        fragment_end_1,
        ["item3"],
        fragment_end_2,
        ["item4"],
    ]


def test_batchify_by_fragment_sentinel_split():
    sentinel = MockStreamSentinel()
    fragment_end = FragmentEndSentinel("fragment")
    data = ["item1", sentinel, "item2", fragment_end]
    batches = list(batchify_by_fragment(data, sentinel_mode="split"))
    assert batches == [["item1"], sentinel, ["item2"], fragment_end]


def test_batchify_by_fragment_sentinel_keep():
    sentinel = MockStreamSentinel()
    fragment_end = FragmentEndSentinel("fragment")
    data = ["item1", sentinel, "item2", fragment_end]
    batches = list(batchify_by_fragment(data, sentinel_mode="keep"))
    assert batches == [["item1", sentinel, "item2"], fragment_end]


def test_batchify_by_fragment_sentinel_drop():
    sentinel = MockStreamSentinel()
    fragment_end = FragmentEndSentinel("fragment")
    data = ["item1", sentinel, "item2", fragment_end]
    batches = list(batchify_by_fragment(data, sentinel_mode="drop"))
    assert batches == [["item1", "item2"], fragment_end]


def test_batchify_by_fragment_drop_last():
    fragment_end = FragmentEndSentinel("fragment")
    data = ["item1", "item2", fragment_end]
    batches = list(batchify_by_fragment(data, sentinel_mode="split", drop_last=True))
    assert batches == [["item1", "item2"], fragment_end]


# Tests for stat_batchify
def test_stat_batchify_simple():
    data = [
        {"/stats/length": 2, "text": "aa"},
        {"/stats/length": 3, "text": "bbb"},
        {"/stats/length": 4, "text": "cccc"},
        {"/stats/length": 2, "text": "dd"},
    ]
    batch_fn = stat_batchify("length")
    batches = list(batch_fn(data, batch_size=5))
    assert batches == [
        [data[0], data[1]],  # Total length: 5
        [data[2]],  # Total length: 4
        [data[3]],  # Total length: 2
    ]


def test_stat_batchify_invalid_key():
    data = [{"text": "aaa"}]
    batch_fn = stat_batchify("length")
    with pytest.raises(ValueError):
        list(batch_fn(data, batch_size=5))


def test_stat_batchify_sentinel_split():
    sentinel = MockStreamSentinel()
    data = [
        {"/stats/length": 2, "text": "aa"},
        sentinel,
        {"/stats/length": 3, "text": "bbb"},
    ]
    batch_fn = stat_batchify("length")
    batches = list(batch_fn(data, batch_size=5, sentinel_mode="split"))
    assert batches == [
        [data[0]],
        sentinel,
        [data[2]],
    ]


def test_stat_batchify_sentinel_keep():
    sentinel = MockStreamSentinel()
    data = [
        {"/stats/length": 2, "text": "aa"},
        sentinel,
        {"/stats/length": 4, "text": "bbbb"},
    ]
    batch_fn = stat_batchify("length")
    batches = list(batch_fn(data, batch_size=5, sentinel_mode="keep"))
    assert batches == [
        [data[0], sentinel],
        [data[2]],
    ]


def test_stat_batchify_drop_last():
    data = [
        {"/stats/length": 2, "text": "aa"},
        {"/stats/length": 3, "text": "bbb"},
        {"/stats/length": 4, "text": "cccc"},
    ]
    batch_fn = stat_batchify("length")
    batches = list(batch_fn(data, batch_size=6, drop_last=True))
    assert batches == [
        [data[0], data[1]],  # Total length: 5
    ]  # Last batch is dropped because total length is 4


# Additional tests to ensure full coverage
def test_batchify_empty_iterable():
    data = []
    batches = list(batchify(data, batch_size=2))
    assert batches == []


def test_batchify_by_length_sum_empty_iterable():
    data = []
    batches = list(batchify_by_length_sum(data, batch_size=5))
    assert batches == []


def test_batchify_by_padded_empty_iterable():
    data = []
    batches = list(batchify_by_padded(data, batch_size=6))
    assert batches == []


def test_batchify_by_dataset_empty_iterable():
    data = []
    batches = list(batchify_by_dataset(data))
    assert batches == []


def test_batchify_by_fragment_empty_iterable():
    data = []
    batches = list(batchify_by_fragment(data))
    assert batches == []


def test_stat_batchify_empty_iterable():
    data = []
    batch_fn = stat_batchify("length")
    batches = list(batch_fn(data, batch_size=5))
    assert batches == []


def test_batchify_invalid_sentinel_mode():
    data = [1, 2, 3]
    with pytest.raises(AssertionError):
        list(batchify(data, batch_size=2, sentinel_mode="invalid_mode"))



from edsnlp.utils.examples import Entity, parse_example

examples = [
    "J'ai vu le patient à cause d'une <ent modality=suicide_attempt_unspecific>TS</ent> médicamenetuse."  # noqa: E501
    "J'ai vu le patient à cause d'une ts médicamenetuse.",  # noqa: E501
    "J'ai vu le patient à cause d'une <ent modality=intentional_drug_overdose>IMV</ent>.",  # noqa: E501
    "surface TS",  # noqa: E501
    "Patiente hospitalisée à cause d'une <ent modality=autolysis>Tentative d'autolyse</ent>.",  # noqa: E501
    "Le patient exprime des idées de défenestration",  # noqa: E501
    "vu aux urgences suite à une <ent modality=jumping_from_height>défenestration volontaire</ent>",  # noqa: E501
    "amené par les pompiers à cause d'une <ent modality=cuts>phlebotomie</ent>",  # noqa: E501
    "Antécédents :\n- <ent modality=strangling>pendaison</ent> (2010)",  # noqa: E501
    "copain du patient : plusieurs événements d'<ent modality=self_destructive_behavior>autodestruction</ent>",  # noqa: E501
    "suspicion d'<ent modality=burn_gas_caustic>ingestion de caustique</ent> avec des idées suicidaires",  # noqa: E501
]


def test_suicide_attempt(blank_nlp):
    blank_nlp.add_pipe("eds.suicide_attempt")

    for text, entities in map(parse_example, examples):
        doc = blank_nlp(text)

        assert len(doc.ents) == len(entities)

        for ent, entity in zip(doc.ents, entities):
            entity: Entity
            assert ent.text == text[entity.start_char : entity.end_char]
            assert ent._.suicide_attempt_modality == entity.modifiers_dict.get(
                "modality"
            )



from __future__ import annotations

import copyreg
import io
import logging
import math
import multiprocessing
import multiprocessing.reduction
import os
import re
import sys
import tempfile
import threading
import time
import traceback
import warnings
import weakref
from collections import defaultdict
from contextlib import nullcontext
from itertools import tee
from multiprocessing.connection import wait
from multiprocessing.queues import Empty
from typing import (
    Dict,
    Iterable,
    List,
    Sequence,
    TypeVar,
    Union,
)

import dill
from tqdm import tqdm

from edsnlp.core.stream import Stage, Stream, StreamSentinel
from edsnlp.data.base import BatchWriter
from edsnlp.reducers import pickler_dont_save_module_dict
from edsnlp.utils.collections import (
    batch_compress_dict,
    decompress_dict,
)

WORKER_STOP_CHECK_INTERVAL = 5.0  # seconds


class StopType:
    # Singleton is important since the STOP objects may be passed to
    # other processes, i.e. pickled, unpickled, while they should
    # always be the same object.
    instance = None

    def __repr__(self):
        return "STOP"

    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__new__(cls)
        return cls.instance


STOP = StopType()


class ForkingPickler(dill.Pickler):
    """
    ForkingPickler that uses dill instead of pickle to transfer objects between
    processes.
    """

    _extra_reducers = {}
    _copyreg_dispatch_table = copyreg.dispatch_table

    def __new__(cls, *args, **kwargs):
        result = dill.Pickler.__new__(ForkingPickler)
        # Python would not call __init__ if the original
        # multiprocessing.reduction.ForkingPickler called, leading to a call to this
        # monkey-patched __new__ method, because [original cls] != [type of result]
        # (see https://docs.python.org/3/reference/datamodel.html#basic-customization)
        # so we force the call to __init__ here
        kwargs = dict(recurse=True, byref=True, **kwargs)
        if not isinstance(result, cls):
            result.__init__(*args, **kwargs)
        return result

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.dispatch_table = self._copyreg_dispatch_table.copy()
        self.dispatch_table.update(self._extra_reducers)

    @classmethod
    def register(cls, type, reduce):
        """Register a reduce function for a type."""
        cls._extra_reducers[type] = reduce

    @classmethod
    def dumps(cls, obj, protocol=None, *args, **kwds):
        buf = io.BytesIO()
        cls(buf, protocol, *args, **kwds).dump(obj)
        return buf.getbuffer()

    loads = dill.loads


def replace_pickler():
    """
    Replace the default pickler used by multiprocessing with dill.
    "multiprocess" didn't work for obscure reasons (maybe the reducers / dispatchers
    are not propagated between multiprocessing and multiprocess => torch specific
    reducers might be missing ?), so this patches multiprocessing directly.
    directly.

    For some reason I do not explain, this has a massive impact on the performance of
    the multiprocessing backend. With the original pickler, the performance can be
    up to 2x slower than with our custom one.
    """
    old_pickler = multiprocessing.reduction.ForkingPickler

    before = (
        dict(ForkingPickler._extra_reducers),
        old_pickler.__new__,
        old_pickler.dumps,
        old_pickler.loads,
        old_pickler.register,
    )

    old_pickler.__new__ = ForkingPickler.__new__
    old_pickler.dumps = ForkingPickler.dumps
    old_pickler.loads = ForkingPickler.loads
    old_pickler.register = ForkingPickler.register
    ForkingPickler._extra_reducers.update(
        multiprocessing.reduction.ForkingPickler._extra_reducers  # noqa
    )

    def revert():
        (
            ForkingPickler._extra_reducers,
            old_pickler.__new__,
            old_pickler.dumps,
            old_pickler.loads,
            old_pickler.register,
        ) = before

    return revert


def cpu_count():  # pragma: no cover
    """
    Heavily inspired (partially copied) from joblib's loky
    (https://github.com/joblib/loky/blob/2c21e/loky/backend/context.py#L83)
    by Thomas Moreau and Olivier Grisel.

    Return the number of CPUs we can use to process data in parallel.

    The returned number of CPUs returns the minimum of:
     * `os.cpu_count()`
     * the CPU affinity settings
     * cgroup CPU bandwidth limit (share of total CPU time allowed in a given job)
       typically used in containerized environments like Docker

    Note that on Windows, the returned number of CPUs cannot exceed 61 (or 60 for
    Python < 3.10), see:
    https://bugs.python.org/issue26903.

    It is also always larger or equal to 1.
    """
    # Note: os.cpu_count() is allowed to return None in its docstring
    os_cpu_count = os.cpu_count() or 1
    if sys.platform == "win32":
        # Following loky's windows implementation

        _MAX_WINDOWS_WORKERS = 60
        if sys.version_info >= (3, 8):
            from concurrent.futures.process import _MAX_WINDOWS_WORKERS  # noqa

            if sys.version_info < (3, 10):
                _MAX_WINDOWS_WORKERS = _MAX_WINDOWS_WORKERS - 1
        os_cpu_count = min(os_cpu_count, _MAX_WINDOWS_WORKERS)

    cpu_count_affinity = os_cpu_count
    try:
        cpu_count_affinity = len(os.sched_getaffinity(0))
    except (NotImplementedError, AttributeError):
        pass

    # Cgroup CPU bandwidth limit available in Linux since 2.6 kernel
    cpu_count_cgroup = os_cpu_count
    cpu_max_fname = "/sys/fs/cgroup/cpu.max"
    cfs_quota_fname = "/sys/fs/cgroup/cpu/cpu.cfs_quota_us"
    cfs_period_fname = "/sys/fs/cgroup/cpu/cpu.cfs_period_us"
    if os.path.exists(cpu_max_fname):
        # cgroup v2
        # https://www.kernel.org/doc/html/latest/admin-guide/cgroup-v2.html
        with open(cpu_max_fname) as fh:
            cpu_quota_us, cpu_period_us = fh.read().strip().split()
    elif os.path.exists(cfs_quota_fname) and os.path.exists(cfs_period_fname):
        # cgroup v1
        # https://www.kernel.org/doc/html/latest/scheduler/sched-bwc.html#management
        with open(cfs_quota_fname) as fh:
            cpu_quota_us = fh.read().strip()
        with open(cfs_period_fname) as fh:
            cpu_period_us = fh.read().strip()
    else:
        cpu_quota_us = "max"
        cpu_period_us = 100_000

    if cpu_quota_us != "max":
        cpu_quota_us = int(cpu_quota_us)
        cpu_period_us = int(cpu_period_us)
        if cpu_quota_us > 0 and cpu_period_us > 0:
            cpu_count_cgroup = math.ceil(cpu_quota_us / cpu_period_us)

    return max(1, min(os_cpu_count, cpu_count_affinity, cpu_count_cgroup))


def hdfs_or_jvm_loaded() -> bool:
    libs = ("libhdfs", "libhdfs3", "libhadoop", "libjvm")

    def loaded_libs():
        if sys.platform.startswith("linux"):
            try:
                with open("/proc/self/maps", "r") as f:
                    return re.findall(r"/[^\s]+\.so[^\s]*", f.read())
            except Exception:
                return []
        try:
            import psutil

            return [m.path for m in psutil.Process().memory_maps() if m.path]
        except Exception:
            return []

    try:
        loaded = loaded_libs()
    except Exception:
        return False
    return any(any(k in os.path.basename(p).lower() for k in libs) for p in loaded)


def get_multiprocessing_context(has_torch_pipes, process_start_method):
    methods = multiprocessing.get_all_start_methods()
    default_method = "fork" if "fork" in methods else "spawn"
    has_hdfs = hdfs_or_jvm_loaded()

    # Base choice, ie torch based pipes prefer spawn
    method = process_start_method or (
        "spawn" if has_torch_pipes or has_hdfs else default_method
    )

    # Warn for torch + fork
    if has_torch_pipes and method == "fork":
        warnings.warn(
            "Using fork start method with GPU workers may lead to deadlocks. "
            "Consider using process_start_method='spawn' instead."
        )
        method = "spawn"

    # Avoid fork if HDFS/JVM already loaded
    if has_hdfs and method == "fork":
        safe = "forkserver" if "forkserver" in methods else "spawn"
        warnings.warn(
            "Using fork start method with HDFS may lead to deadlocks. "
            f"Consider using process_start_method='{safe}' instead."
        )
        method = safe

    if default_method != method:
        logging.info(f"Switching process start method to {method}")

    return multiprocessing.get_context(method)


# Should we check if the multiprocessing module of edsnlp
# is responsible for this child process before replacing the pickler ?
if (
    multiprocessing.current_process() != "MainProcess"
    or hasattr(multiprocessing, "parent_process")
    and multiprocessing.parent_process() is not None
):
    replace_pickler()


try:
    import torch.multiprocessing

    from edsnlp.utils.torch import dump, load

except (ImportError, AttributeError):  # pragma: no cover
    # noinspection PyUnusedLocal
    def load(file, *args, map_location=None, **kwargs):
        # check if path
        if isinstance(file, str):
            with open(file, "rb") as f:
                return dill.load(f, *args, **kwargs)
        return dill.load(file, *args, **kwargs)

    def dump(obj, file, skip_tensors=False, *args, **kwargs):
        old_settings = dict(dill.settings)
        try:
            dill.settings["recurse"] = False
            dill.settings["byref"] = True
            return dill.dump(obj, file, *args, **kwargs)
        finally:
            dill.settings.update(old_settings)


if os.environ.get("TORCH_SHARING_STRATEGY"):  # pragma: no cover
    try:
        torch.multiprocessing.set_sharing_strategy(os.environ["TORCH_SHARING_STRATEGY"])
    except NameError:
        pass

U = TypeVar("U")
T = TypeVar("T")


def get_dispatch_schedule(
    producer_idx: U,
    producers: Sequence[U],
    consumers: Sequence[T],
) -> List[T]:
    """
    To which consumer should a given worker/producer dispatch its data to. This
    function returns a list of consumers over a period to be determined by the
    function.

    This is actually a fun problem, because we want:
    - to distribute the data evenly between consumers
    - minimize the number of distinct unique producers sending data to the consumer
      (because we move the tensors to the GPU inside the producer, which
       creates takes a bit of VRAM for each producer)

    Parameters
    ----------
    producer_idx: U
        Index of the CPU worker
    producers: Sequence[U]
        Producers, ie workers
    consumers: Sequence[T]
        Consumers, ie devices

    Returns
    -------
    List[T]
    """
    if not (isinstance(consumers, range) and isinstance(producers, range)):
        p_range = range(len(producers))
        c_range = range(len(consumers))
        schedule = get_dispatch_schedule(
            producers.index(producer_idx), p_range, c_range
        )
        return [consumers[i] for i in schedule]

    base = []

    if consumers:
        if len(consumers) < len(producers):
            r = len(producers) % len(
                consumers
            )  # number of CPU without an easy to assign GPU
            R = producers[len(producers) - r :]
            if producer_idx not in R:
                idx = producers.index(producer_idx)
                base = [consumers[idx % len(consumers)]]
                base = base * (len(consumers) // len(base))
            else:
                base = get_dispatch_schedule(producer_idx, R, consumers)
        else:
            idx = producers.index(producer_idx)
            r = len(consumers) % len(producers)
            d = len(consumers) // len(producers)
            base = list(consumers[idx * d : (idx + 1) * d])
            base = base * len(producers)
            R = consumers[-r:]
            if r > 0:
                base = base + get_dispatch_schedule(producer_idx, producers, R)
    return base


class StopSignal(BaseException):
    pass


class Worker:
    def __init__(
        self,
        uid: str,
        num_cpu_workers: int,
        data_queues: Dict[str, multiprocessing.Queue],
        main_control_queue: multiprocessing.Queue,
        worker_control_queue: multiprocessing.Queue,
        final_barrier: multiprocessing.Barrier,
        stream_path: str,
        devices: Dict[str, Union[str, "torch.device"]],
        gpu_semaphores: Dict[str, multiprocessing.Semaphore],
    ):
        super(Worker, self).__init__()

        self.uid = uid
        self.num_cpu_workers = num_cpu_workers
        self.stream_path = stream_path
        self.stream: Stream = None  # will be assigned in run
        self.stages: List[Stage] = None  # will be assigned in run
        self.devices = devices
        self.num_producers_alive = {}
        self.stop = False

        # Synchronization primitives
        self.data_queues = data_queues
        self.gpu_semaphores = gpu_semaphores
        self.main_control_queue = main_control_queue
        self.worker_control_queue = worker_control_queue
        self.final_barrier = final_barrier
        self.waiting_times = defaultdict(float)

    def on_stop(self):
        pass

    @property
    def stages_to_run(self):
        raise NotImplementedError()

    def consumer_queues(self, stage):
        raise NotImplementedError()

    def process_items(self, stage):
        raise NotImplementedError()

    def iter_tasks(self, stage, stop_mode=False):
        raise NotImplementedError()

    def run_stage_thread(self, stage):
        self.num_producers_alive[stage] = len(
            [
                name
                for name in self.data_queues
                if name.endswith(f"to-stage-{stage}_of-{self.uid}")
            ]
        )
        try:
            self.process_items(stage)
        except StopSignal:  # pragma: no cover
            if not self.stop:
                self.stop = True
                self.on_stop()
        except BaseException as e:
            if self.stop:  # pragma: no cover
                return
            print(f"Error in {self.uid}/{stage}:\n{traceback.format_exc()}", flush=True)
            self.main_control_queue.put(e)
        finally:
            try:
                for _ in self.iter_tasks(stage=stage, stop_mode=True):
                    pass
            except StopSignal:
                pass
            for name, queue in self.consumer_queues(stage):
                queue.put(STOP)

    def run(self):
        threads = []
        try:
            device = self.devices.get(self.uid, "cpu")
            self.stream, self.stages = load(self.stream_path, map_location=device)
            self.stream.eval()
            stages_to_run = self.stages_to_run

            for stage_idx in stages_to_run:
                thread = threading.Thread(
                    target=self.run_stage_thread,
                    args=(stage_idx,),
                    daemon=True,
                    name=f"Worker-{self.uid}_stage-{stage_idx}",
                )
                threads.append(thread)
                thread.start()

            # Inform the main process that we are ready
            self.main_control_queue.put("READY")

            while not self.stop:
                try:
                    notification = self.worker_control_queue.get(
                        timeout=WORKER_STOP_CHECK_INTERVAL
                    )
                except Empty:  # pragma: no cover
                    continue
                if notification is STOP and not self.stop:
                    self.stop = True
                    self.on_stop()

        except BaseException as e:  # pragma: no cover
            if self.stop:
                return
            self.stop = True
            self.on_stop()

            # KeyboardInterrupt occurs in every process, so we don't want to print
            # it everywhere since it will be printed in the main process
            if not isinstance(e, KeyboardInterrupt):
                print(f"Error in {self.uid}:\n{traceback.format_exc()}", flush=True)
            self.main_control_queue.put(e)
        finally:
            for thread in threads:
                thread.join()
            for stage in self.stages_to_run:
                for name, queue in self.consumer_queues(stage):
                    queue.close()
                    if hasattr(queue, "join_thread"):
                        queue.join_thread()

            self.final_barrier.wait()


class CPUWorker(Worker):
    def __init__(self, *args, schedule, **kwargs):
        super().__init__(*args, **kwargs)
        self.active_batches = {}
        self.batch_devices = {}
        self.schedule = schedule
        self.stage_schedule = defaultdict(lambda: list(schedule))

    def on_stop(self):
        # Callback when stopping
        for name, sem in self.gpu_semaphores.items():
            # just unlock everything at this point
            # Ideally we shouldn't need this and let the threads finish
            # normally, but since we don't send data anymore to GPU when
            # stop is True, logically we won't receive the results from
            # the GPU and if the semaphore was just acquired before stop
            # was set to True, another thread can be stuck waiting for it
            # without any chance for the semaphore to be released.
            # So we just unlock everything here.
            sem.release()

    def process_items(self, stage):
        items = self.iter_tasks(stage)

        # If we have a torch pipe in the previous stage, we may need to postprocess the
        # results before applying subsequent ops
        if stage > 0:
            items = self.postprocess_after_forward(items, stage)

        # Apply the ops
        for op in self.stages[stage].cpu_ops:
            items = op(items)

        # If we have a torch pipe in the current stage, we need to preprocess the items
        # before sending them to the GPU workers and apply the forward pass
        if self.stages[stage].gpu_op is not None:
            self.preprocess_before_forward(items, stage)
        else:
            self.send_results(items)

    def iter_tasks(self, stage, stop_mode=False):
        if self.stream.reader.read_in_worker and stage == 0:
            local_rank = int(os.environ.get("LOCAL_RANK", 0))
            world_size = int(os.environ.get("WORLD_SIZE", 1))
            worker_idx = int(self.uid[3:]) + self.num_cpu_workers * local_rank
            pool_size = self.num_cpu_workers * world_size

            if stop_mode:
                return
            task_idx = 0
            for item in iter(self.stream.reader.read_records()):
                if self.stop:  # pragma: no cover
                    # I don't know why, sometimes when a KeyboardInterrupt is captured
                    # by the main process, and converted to a stop=True in the workers,
                    # this line is correctly reached, but the StopSignal exception is
                    # not bubbled up most of the time to `self.send_results(items)` in
                    # `CPUWorker.process_items` and therefore not caught in
                    # `self.process_items(stage)` in `Worker.run_stage_thread`.
                    # This makes the worker hang at the end instead of stopping.
                    # Simply wrapping this iter_tasks() generator in a try/except
                    # next(iterator) doesn't catch the exception when raised here, as if
                    # the exception bubbling mechanism was broken in these cases.
                    raise StopSignal()
                if isinstance(item, StreamSentinel):
                    yield item
                    continue
                if task_idx % pool_size == worker_idx:
                    for item in self.stream.reader.extract_task(item):
                        if self.stop:  # pragma: no cover
                            raise StopSignal()
                        yield item

                task_idx += 1
        else:
            last_task_idx = -1
            deterministic = self.stream.deterministic
            schedule = self.stage_schedule[stage]

            while self.num_producers_alive[stage] > 0:
                if self.stop and not stop_mode:  # pragma: no cover
                    raise StopSignal()

                if stage > 0:
                    task_idx = (last_task_idx + 1) % len(schedule)
                    while schedule[task_idx] is None:  # pragma: no cover
                        task_idx = (task_idx + 1) % len(schedule)

                    if deterministic:
                        prod = schedule[task_idx]
                    else:
                        rolled = schedule[task_idx:] + schedule[:task_idx]
                        producers = list(
                            dict.fromkeys(p for p in rolled if p is not None)
                        )
                        conns = [
                            self.data_queues[
                                f"from-{prod}_to-stage-{stage}_of-{self.uid}"
                            ]._reader
                            for prod in producers
                        ]
                        ready = wait(conns)[0]
                        prod = producers[conns.index(ready)]
                else:
                    prod = "main"

                name = f"from-{prod}_to-stage-{stage}_of-{self.uid}"
                queue = self.data_queues[name]
                t = time.time()
                try:
                    item = queue.get(timeout=WORKER_STOP_CHECK_INTERVAL)
                except Empty:  # pragma: no cover
                    continue
                else:
                    if stage > 0:
                        last_task_idx = task_idx
                    self.waiting_times["get-" + name] += time.time() - t

                if item is STOP:
                    schedule[:] = [s if s != prod else None for s in schedule]
                    self.num_producers_alive[stage] -= 1
                    continue
                yield item

    def postprocess_after_forward(self, items, stage):
        postprocess = getattr(self.stages[stage - 1].gpu_op, "postprocess", None)
        deterministic = self.stream.deterministic
        for batch_id, result in items:
            x = self.active_batches[stage - 1]
            slot = next(s for s in x if s[0] == batch_id)
            result = {
                k: v.to("cpu") if hasattr(v, "to") else v for k, v in result.items()
            }
            _, docs, inputs, _ = slot
            if postprocess:
                docs = postprocess(docs, result, inputs)
            else:
                docs = result
            slot[3] = docs
            if deterministic:
                while x and x[0][3] is not None:
                    docs = x.pop(0)[3]
                    yield docs
            else:
                x.remove(slot)
                yield docs

            del docs, result, inputs, slot, x

    def preprocess_before_forward(self, items, stage):
        if stage not in self.active_batches:
            self.active_batches[stage] = []
        task_idx = 0
        torch_pipe = self.stages[stage].gpu_op
        for docs in items:
            if isinstance(docs, StreamSentinel):
                self.active_batches[stage].append([None, None, None, docs])
                continue
            batch_id = str(hash(tuple(id(x) for x in docs)))[-8:] + "-" + self.uid
            torch_pipe.enable_cache(batch_id)
            if stage == 0:
                gpu_idx = self.schedule[task_idx % len(self.schedule)]
                self.batch_devices[batch_id] = gpu_idx
            else:
                # Ensure we send the batch to the same device as the previous stages.
                # In deterministic mode, using schedule[task_idx % len(schedule)]
                # we shouldn't need this, but in non-deterministic mode, the order
                # in which batches are returned to cpu workers may change depending
                # on the speed at which gpu workers process them, rendering
                # the above gpu_idx computation inconsistent over time.
                gpu_idx = self.batch_devices[batch_id]
            task_idx += 1
            device = self.devices[gpu_idx]
            if hasattr(torch_pipe, "preprocess"):
                inputs = [torch_pipe.preprocess(x) for x in docs]
                batch = decompress_dict(list(batch_compress_dict(inputs)))
                batch = torch_pipe.collate(batch)
                batch = torch_pipe.batch_to_device(batch, device=device)
            else:
                batch = torch_pipe.prepare_batch(docs, device=device)
                inputs = None

            self.active_batches[stage].append([batch_id, docs, inputs, None])

            name = f"from-{self.uid}_to-stage-{stage}_of-{gpu_idx}"
            queue = self.data_queues[name]
            item = (self.uid, batch_id, batch)
            if not self.stop and stage == 0:
                sem = self.gpu_semaphores[gpu_idx]
                t = time.time()
                sem.acquire()

                self.waiting_times[f"semaphore-{gpu_idx}"] += time.time() - t
            t = time.time()
            queue.put(item)

            self.waiting_times["put-" + name] += time.time() - t

            del batch, inputs

            if stage == len(self.stages) - 2:
                torch_pipe.disable_cache(batch_id)
                self.batch_devices.pop(batch_id, None)

    def send_results(self, items):
        writer = self.stream.writer
        if writer is not None:
            items = (
                writer.handle_record(rec)
                if not isinstance(rec, StreamSentinel)
                else rec
                for rec in items
            )
        if getattr(writer, "write_in_worker", None) is True:
            items = writer.batch_fn(
                items,
                batch_size=writer.batch_size,
                sentinel_mode="drop",
            )
            items = (
                writer.handle_batch(b)
                for b in items
                if not isinstance(b, StreamSentinel)
            )
        else:
            items = (
                (x, 1) if not isinstance(x, StreamSentinel) else (x, 0) for x in items
            )

        name = f"from-{self.uid}_to-main"
        queue = self.data_queues[name]
        for item in items:
            t = time.time()
            queue.put(item)

            self.waiting_times["put-" + name] += time.time() - t

    @property
    def stages_to_run(self):
        return range(len(self.stages))

    def consumer_queues(self, stage):
        if stage < len(self.stages) - 1:
            names = [
                f"from-{self.uid}_to-stage-{stage}_of-{gpu}"
                for gpu in set(self.schedule)
            ]
            return [(name, self.data_queues[name]) for name in names]
        else:
            name = f"from-{self.uid}_to-main"
            return [(name, self.data_queues[name])]


class GPUWorker(Worker):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lock = threading.Lock()

    def process_items(self, stage):
        autocast = self.stream.autocast
        autocast_ctx = nullcontext()
        device = self.devices[self.uid]
        device_type = getattr(device, "type", device).split(":")[0]
        try:
            if autocast:
                autocast_ctx = torch.autocast(
                    device_type=device_type,
                    dtype=autocast if autocast is not True else None,
                )
        except RuntimeError:  # pragma: no cover
            pass

        with torch.no_grad(), autocast_ctx, torch.inference_mode():
            for item in self.iter_tasks(stage):
                with self.lock:
                    cpu_id, batch_id, batch = item

                    pipe = self.stages[stage].gpu_op
                    pipe.enable_cache(batch_id)
                    batch = pipe(batch)

                    # Put item into the queue
                    name = f"from-{self.uid}_to-stage-{stage + 1}_of-{cpu_id}"
                    queue = self.data_queues[name]
                    item = (batch_id, batch)

                # Do NOT put during lock, otherwise this may lead to a deadlock
                # in multi-stage (n + 1 where n > 1) scenarios where stage 1
                # must put in a queue that is full, and could only be emptied
                # if subsequent stages run, but forward in stage 2 cannot run
                # since we have acquired the lock.

                if stage == len(self.stages) - 2:  # Last GPU stage
                    pipe.disable_cache(batch_id)
                    self.gpu_semaphores[cpu_id].release()
                t = time.time()
                queue.put(item)

                self.waiting_times["put-" + name] += time.time() - t

                del batch, item

    def iter_tasks(self, stage, stop_mode=False):
        queues = [
            key
            for key, q in self.data_queues.items()
            if key.endswith(f"-{stage}_of-{self.uid}")
        ]
        # Get items from the previous stage
        last_offset = -1
        while self.num_producers_alive[stage] > 0:
            if self.stop and not stop_mode:  # pragma: no cover
                raise StopSignal()

            offset = (last_offset + 1) % len(queues)
            while queues[offset] is None:  # pragma: no cover
                offset = (offset + 1) % len(queues)

            # Roll the queues to ensure we don't always get from the same queue
            names = [q for q in queues[offset:] + queues[:offset] if q is not None]
            conns = [self.data_queues[name]._reader for name in names]
            ready = wait(conns)[0]
            name = names[conns.index(ready)]
            queue = self.data_queues[name]
            t = time.time()
            try:
                item = queue.get(timeout=WORKER_STOP_CHECK_INTERVAL)
            except Empty:  # pragma: no cover
                continue
            else:
                last_offset = offset
                self.waiting_times["get-" + name] += time.time() - t

            if item is STOP:
                queues[:] = [s if s != name else None for s in queues]
                self.num_producers_alive[stage] -= 1
                continue
            yield item

    @property
    def stages_to_run(self):
        return range(len(self.stages) - 1)

    def consumer_queues(self, stage):
        names = [
            k
            for k, q in self.data_queues.items()
            if k.startswith(f"from-{self.uid}_to-stage-{stage + 1}_of-")
        ]
        return [(name, self.data_queues[name]) for name in names]


class SafeNext:
    def __init__(self, it, lock):
        self.it = it
        self.lock = lock

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.it)


def thread_safe_tee(it: Iterable[T], n: int):
    lock = threading.Lock()
    return tuple(SafeNext(iter(sub_it), lock) for sub_it in tee(it, n))


class MultiprocessingStreamExecutor:
    def __init__(self, stream):
        (
            num_cpu_workers,
            num_gpu_workers,
            cpu_worker_devices,
            gpu_worker_devices,
            has_torch_pipes,
        ) = self.adjust_num_workers(stream)
        self.stream = stream
        self.stages = stream._make_stages(split_torch_pipes=num_gpu_workers > 0)
        self.has_torch_pipes = has_torch_pipes
        mp = get_multiprocessing_context(
            has_torch_pipes=has_torch_pipes,
            process_start_method=stream.process_start_method,
        )

        # Queues definition
        share_queues = not stream.deterministic

        self.cpu_worker_names = [f"cpu{i}" for i in range(num_cpu_workers)]
        self.gpu_worker_names = [f"gpu{i}" for i in range(num_gpu_workers)]
        self.all_worker_names = self.cpu_worker_names + self.gpu_worker_names

        self.cpu_to_gpu_schedules = {
            cpu: get_dispatch_schedule(
                cpu,
                self.cpu_worker_names,
                self.gpu_worker_names,
            )
            for cpu in self.cpu_worker_names
        }

        self.main_control_queue = mp.Queue()
        self.worker_control_queues = {
            name: mp.Queue() for name in self.all_worker_names
        }
        self.final_barrier = mp.Barrier(num_cpu_workers + num_gpu_workers + 1)
        self.data_queues: Dict[str, multiprocessing.Queue] = {}
        self.gpu_semaphores = {}
        self.input_queue_names = []

        # Queues: for each N producers - 1 consumer situation, we don't create
        # one single shared queue but N bounded queues, one for each producer.
        # This is to avoid the situation where a single producer occupies all
        # slots, leading to congestions affecting the whole workers pool.

        # Input queues for each CPU worker
        if not self.stream.reader.read_in_worker:
            queue = mp.Queue(2 * num_cpu_workers) if share_queues else None
            for cpu in self.cpu_worker_names:
                name = f"from-main_to-stage-0_of-{cpu}"
                if not share_queues:
                    queue = mp.Queue()
                self.data_queues[name] = queue
                self.input_queue_names.append(name)

        for cpu in set(self.cpu_worker_names):
            for gpu in set(self.cpu_to_gpu_schedules[cpu]):
                # Control the number of active items for each CPU -> GPU pair
                self.gpu_semaphores[(cpu, gpu)] = mp.Semaphore(2)

                for stage in range(0, len(self.stages) - 1):
                    # Queue to send data from CPU to GPU
                    name = f"from-{cpu}_to-stage-{stage}_of-{gpu}"
                    self.data_queues[name] = mp.Queue()

                    # Answer queue from GPU to CPU
                    name = f"from-{gpu}_to-stage-{stage + 1}_of-{cpu}"
                    self.data_queues[name] = mp.Queue()

            # Final output queue for each CPU worker
            name = f"from-{cpu}_to-main"
            self.data_queues[name] = mp.Queue(2)

        self.cpu_temp_file = self.gpu_temp_file = None
        if len(self.cpu_worker_names):
            self.cpu_temp_file = tempfile.NamedTemporaryFile(delete=False)
        if len(self.gpu_worker_names):
            self.gpu_temp_file = tempfile.NamedTemporaryFile(delete=False)

        self.cpu_workers = []
        self.gpu_workers = []
        devices = {
            **(dict(zip(self.cpu_worker_names, cpu_worker_devices))),
            **(dict(zip(self.gpu_worker_names, gpu_worker_devices))),
        }
        for cpu in self.cpu_worker_names:
            worker = CPUWorker(
                uid=cpu,
                num_cpu_workers=num_cpu_workers,
                data_queues={k: q for k, q in self.data_queues.items() if cpu in k},
                main_control_queue=self.main_control_queue,
                worker_control_queue=self.worker_control_queues[cpu],
                final_barrier=self.final_barrier,
                schedule=self.cpu_to_gpu_schedules[cpu],
                stream_path=self.cpu_temp_file.name,
                devices=devices,
                gpu_semaphores={
                    gpu: sem
                    for (c, gpu), sem in self.gpu_semaphores.items()
                    if cpu == c
                },
            )
            self.cpu_workers.append(
                mp.Process(
                    target=worker.run,
                    name=f"CPUWorker-{cpu}",
                    daemon=True,
                )
            )
        for gpu in self.gpu_worker_names:
            worker = GPUWorker(
                uid=gpu,
                num_cpu_workers=num_cpu_workers,
                data_queues={k: q for k, q in self.data_queues.items() if gpu in k},
                main_control_queue=self.main_control_queue,
                worker_control_queue=self.worker_control_queues[gpu],
                final_barrier=self.final_barrier,
                stream_path=self.gpu_temp_file.name,
                devices=devices,
                gpu_semaphores={
                    cpu: sem
                    for (cpu, g), sem in self.gpu_semaphores.items()
                    if gpu == g
                },
            )
            self.gpu_workers.append(
                mp.Process(
                    target=worker.run,
                    name=f"GPUWorker-{gpu}",
                    daemon=True,
                )
            )
        self.stopped = False
        self.num_alive_workers = num_cpu_workers
        self.workers_status = [True] * num_cpu_workers
        self.current_worker_idx = 0
        self.error = None
        self.dequeue_notifications_thread = None
        self.queue_feeder_threads: List[threading.Thread] = []
        self.revert_environ = lambda: None
        self.revert_pickler = lambda: None
        self.tearing_down = False

        logging.debug(f"Main PID {os.getpid()}")
        logging.info(
            f"Running {num_cpu_workers} CPU workers and {num_gpu_workers} GPU "
            f"workers on {gpu_worker_devices} in {mp.get_start_method()} mode "
            f"to run {len(self.stages)} stage{'s' if len(self.stages) > 1 else ''}.",
        )

    def run(self):
        self.revert_pickler = replace_pickler()
        self.revert_environ = self.setup_environ(
            self.stream.disable_implicit_parallelism
        )

        stream_to_dump = self.stream.worker_copy()
        with pickler_dont_save_module_dict():
            if self.cpu_temp_file:
                # If we have GPU workers, these will be responsible for the forward pass
                # and CPU workers will only perform preprocessing and postprocessing
                # so they don't need deep learning parameters
                # TODO: should we make this a stream set_processing option ?
                keep_tensors = self.has_torch_pipes and len(self.gpu_worker_names) == 0
                with self.cpu_temp_file as fp:
                    dump(
                        (stream_to_dump, self.stages),
                        fp,
                        skip_tensors=not keep_tensors,
                    )
                    fp.close()
            if self.gpu_temp_file:
                with self.gpu_temp_file as fp:
                    dump((stream_to_dump, self.stages), fp)
                    fp.close()

        del stream_to_dump
        for worker in (*self.cpu_workers, *self.gpu_workers):
            worker.start()

        # Wait for all workers to be ready before deleting the temp file
        # and starting the main loop
        for _ in range(len(self.cpu_workers) + len(self.gpu_workers)):
            outputs = self.main_control_queue.get()
            if isinstance(outputs, BaseException):  # pragma: no cover
                raise outputs

        if self.cpu_temp_file is not None:
            os.unlink(self.cpu_temp_file.name)
        if self.gpu_temp_file is not None:
            os.unlink(self.gpu_temp_file.name)

        # Start listening for notifications from workers
        self.dequeue_notifications_thread = threading.Thread(
            target=self.dequeue_notifications,
            name="Main-Notifications",
            daemon=True,
        )
        self.dequeue_notifications_thread.start()

        # Start enqueuing inputs if needed
        if not self.stream.reader.read_in_worker:
            queues = {self.data_queues[name] for name in self.input_queue_names}
            tee_items = thread_safe_tee(self.stream.reader.read_records(), len(queues))
            for queue, items in zip(queues, tee_items):
                thread = threading.Thread(
                    target=self.feed_queue,
                    name="Main-Enqueue-Inputs",
                    daemon=True,
                    args=(queue, items),
                )
                thread.start()
                self.queue_feeder_threads.append(thread)

        # Create the main iterator
        items = self.dequeue_outputs()
        writer = self.stream.writer
        if getattr(writer, "write_in_worker", None) is False:
            writer: BatchWriter
            items = writer.batch_fn(items, writer.batch_size, sentinel_mode="drop")
            # get the 1st element (2nd is the count)
            items = (
                writer.handle_batch(b)[0]
                for b in items
                if not isinstance(b, StreamSentinel)
            )

        # If we are garbage collected, stop the execution
        weakref.finalize(items, self.teardown, garbage_collected=True)

        if writer is not None:
            items = writer.consolidate(items)

        return items

    def dequeue_outputs(self):
        try:
            bar = tqdm(
                smoothing=0.1,
                mininterval=1.0,
                disable=not self.stream.show_progress,
            )
            with bar:
                for item, count in self.iter_outputs():
                    bar.update(count)
                    yield item
        except StopSignal:
            if self.error:
                raise self.error
        except BaseException as e:
            self.error = e
            raise
        finally:
            self.teardown()

    def iter_outputs(self, stop_mode=False):
        deterministic = self.stream.deterministic
        requires_sentinel = (
            hasattr(self.stream.writer, "batch_fn")
            and getattr(self.stream.writer.batch_fn, "requires_sentinel", None)
            and not self.stream.writer.write_in_worker
        )
        missing_sentinels = len(self.cpu_worker_names) if requires_sentinel else 0
        buffer = []
        last_heartbeat = {name: time.time() for name in self.cpu_worker_names}
        while self.num_alive_workers > 0:
            if self.stopped and not stop_mode:  # pragma: no cover
                raise StopSignal()

            if deterministic:
                worker_idx = self.current_worker_idx
            else:
                workers = [
                    *range(self.current_worker_idx, len(self.cpu_worker_names)),
                    *range(self.current_worker_idx),
                ]
                workers = [idx for idx in workers if self.workers_status[idx]]
                output_conns = [
                    self.data_queues[
                        f"from-{self.cpu_worker_names[idx]}_to-main"
                    ]._reader
                    for idx in workers
                ]
                pipe = wait(output_conns)[0]
                worker_idx = workers[output_conns.index(pipe)]

            name = f"from-{self.cpu_worker_names[worker_idx]}_to-main"
            queue = self.data_queues[name]

            try:
                out = queue.get(timeout=10)
            except Empty:  # pragma: no cover
                continue

            last_heartbeat[self.cpu_worker_names[worker_idx]] = time.time()
            self.current_worker_idx = (self.current_worker_idx + 1) % len(
                self.cpu_worker_names
            )
            while not self.workers_status[self.current_worker_idx]:
                self.current_worker_idx = (self.current_worker_idx + 1) % len(
                    self.cpu_worker_names
                )

            if out is STOP:
                self.num_alive_workers -= 1
                self.workers_status[worker_idx] = False
                continue
            if isinstance(out[0], StreamSentinel):
                if out[0].kind == requires_sentinel:
                    missing_sentinels -= 1
                    if missing_sentinels == 0:
                        yield from buffer
                        yield out
                        buffer.clear()
                        missing_sentinels = len(self.cpu_worker_names)
                continue
            if requires_sentinel:
                buffer.append(out)
            else:
                yield out
        yield from buffer

    def feed_queue(self, queue, items):
        """
        Enqueue items in a queue.
        Note that a queue may be shared between multiple workers, so have to send
        items destined multiple workers in the same queue. For that, we first
        determine which worker should receive the item based on the item index
        and some other env variables. Then we lookup the worker queue, and if it
        matches the current queue, we send the item, even if all workers share the
        same queue, in which case there is only on queue feeder thread that sends
        all the items (non-deterministic mode).

        Parameters
        ----------
        queue: multiprocessing.Queue
            The queue to feed
        items: Iterator
            The items to send. Note that this iterator is a tee of the main
            iterator, such that each worker can process items at its own pace.
        """
        queues = [
            self.data_queues[f"from-main_to-stage-0_of-{cpu}"]
            for cpu in self.cpu_worker_names
        ]
        try:
            local_rank = int(os.environ.get("LOCAL_RANK", 0))
            world_size = int(os.environ.get("WORLD_SIZE", 1))

            task_idx = 0
            for item in items:
                if self.stopped:
                    break
                if isinstance(item, StreamSentinel):
                    queue.put(item)
                    continue
                # tasks:         [0, 1, 2, 3, 4, 5, 6, 7, 8, 9 ...]
                # world_size = 2
                # local_rank = 1
                # -> kept tasks: [   1,    3,    5,    7,    9, ...]
                # worker_idx = 1
                # num_cpu_workers = 3
                # -> kept tasks: [         3,                9, ...]
                if (
                    # check that this task is for us
                    (task_idx % world_size) == local_rank
                    # check that this task is for the queue
                    and queues[(task_idx // world_size) % len(queues)] is queue
                ):
                    queue.put(item)
                task_idx += 1
        except BaseException as e:
            print(f"Error in input enqueueing:\n{traceback.format_exc()}", flush=True)
            self.main_control_queue.put(e)
        finally:
            # Send the stop sentinel to all workers
            for q in queues:
                if q is queue:
                    queue.put(STOP)
            if hasattr(queue, "close"):
                queue.close()
            if hasattr(queue, "join_thread"):
                queue.join_thread()

    def dequeue_notifications(self):
        while True:
            notification = self.main_control_queue.get()
            if notification is STOP:
                break
            if isinstance(notification, BaseException):
                self.error = notification
                self.send_stop_signals()

    def teardown(self, garbage_collected=False):
        if self.tearing_down:
            return
        self.tearing_down = True

        if self.error:
            warnings.warn(
                "An error occurred. Cleaning up resources, please hang tight...",
            )
        elif garbage_collected:
            warnings.warn(
                "Multiprocessing executor was garbage collected while still running. "
                "Cleaning up resources, please hang tight...",
            )

        self.send_stop_signals()

        self.revert_environ()
        self.revert_pickler()

        self.error = None
        for _ in self.iter_outputs(stop_mode=True):
            pass

        if self.dequeue_notifications_thread is not None:
            self.dequeue_notifications_thread.join()
        for thread in self.queue_feeder_threads:
            thread.join()
        self.final_barrier.wait()

    def send_stop_signals(self):
        if self.stopped:
            return

        self.stopped = True

        for uid in (*self.cpu_worker_names, *self.gpu_worker_names):
            self.worker_control_queues[uid].put(STOP)
        self.main_control_queue.put(STOP)

    @staticmethod
    def adjust_num_workers(stream: Stream):
        num_gpu_workers = (
            stream.num_gpu_workers
            if stream.num_gpu_workers is not None or stream.gpu_worker_devices is None
            else len(stream.gpu_worker_devices)
        )
        has_torch_pipes = any(stream.torch_components())
        requires_gpu_workers = has_torch_pipes and (
            num_gpu_workers is None
            or num_gpu_workers is not None
            and num_gpu_workers > 0
        )
        num_cpus = int(os.environ.get("EDSNLP_MAX_CPU_WORKERS") or cpu_count())
        num_devices = 0
        if requires_gpu_workers:
            import torch

            num_devices = torch.cuda.device_count()
            logging.debug(f"Number of available devices: {num_devices}")

            if num_gpu_workers is None:
                num_gpu_workers = min(num_devices, num_cpus // 2)
        else:
            num_gpu_workers = 0

        max_cpu_workers = max(num_cpus - num_gpu_workers - 1, 0)
        default_cpu_workers = max(
            min(max_cpu_workers, num_gpu_workers * 4)
            if num_gpu_workers > 0
            else max_cpu_workers,
            1,
        )
        num_cpu_workers = (
            default_cpu_workers
            if stream.num_cpu_workers is None
            else max_cpu_workers + stream.num_cpu_workers + 1
            if stream.num_cpu_workers < 0
            else stream.num_cpu_workers
        )

        gpu_worker_devices = (
            (
                [
                    f"cuda:{gpu_idx * num_devices // num_gpu_workers}"
                    for gpu_idx in range(num_gpu_workers)
                ]
                if stream.gpu_worker_devices is None
                else stream.gpu_worker_devices
            )
            if requires_gpu_workers
            else []
        )
        cpu_worker_devices = (
            ["cpu"] * num_cpu_workers
            if stream.cpu_worker_devices is None
            else stream.cpu_worker_devices
        )
        assert len(cpu_worker_devices) == num_cpu_workers
        assert len(gpu_worker_devices) == num_gpu_workers
        if num_cpu_workers == 0:  # pragma: no cover
            (
                num_cpu_workers,
                num_gpu_workers,
                cpu_worker_devices,
                gpu_worker_devices,
            ) = (num_gpu_workers, 0, gpu_worker_devices, [])

        if num_cpu_workers == 0:
            raise ValueError("At least one worker is required in multiprocessing mode.")

        num_cpu_workers: int
        num_gpu_workers: int
        cpu_worker_devices: List[str]
        gpu_worker_devices: List[str]
        has_torch_pipes: bool
        return (
            num_cpu_workers,
            num_gpu_workers,
            cpu_worker_devices,
            gpu_worker_devices,
            has_torch_pipes,
        )

    @staticmethod
    def setup_environ(disable_implicit_parallelism):
        old_environ = {
            k: os.environ[k]
            for k in (
                "TOKENIZERS_PARALLELISM",
                "OMP_NUM_THREADS",
                "TORCH_SHARING_STRATEGY",
            )
            if k in os.environ
        }
        if disable_implicit_parallelism:
            os.environ["TOKENIZERS_PARALLELISM"] = "false"
            os.environ["OMP_NUM_THREADS"] = "1"

        if "torch" in sys.modules:
            try:
                import torch.multiprocessing

                os.environ["TORCH_SHARING_STRATEGY"] = (
                    torch.multiprocessing.get_sharing_strategy()
                )
            except ImportError:  # pragma: no cover
                pass

        def revert():
            os.environ.update(old_environ)

        return revert


def execute_multiprocessing_backend(stream: Stream):
    """
    If you have multiple CPU cores, and optionally multiple GPUs, we provide the
    `multiprocessing` backend that allows to run the inference on multiple
    processes.

    This accelerator dispatches the batches between multiple workers
    (data-parallelism), and distribute the computation of a given batch on one or two
    workers (model-parallelism):

    - a `CPUWorker` which handles the non deep-learning components and the
      preprocessing, collating and postprocessing of deep-learning components
    - a `GPUWorker` which handles the forward call of the deep-learning components

    If no GPU is available, no `GPUWorker` is started, and the `CPUWorkers` handle
    the forward call of the deep-learning components as well.

    The advantage of dedicating a worker to the deep-learning components is that it
    allows to prepare multiple batches in parallel in multiple `CPUWorker`, and ensure
    that the `GPUWorker` never wait for a batch to be ready.

    The overall architecture described in the following figure, for 3 CPU workers and 2
    GPU workers.

    <div style="text-align:center">
        <img src="/assets/images/multiprocessing.png" style="height:400px" />
    </div>

    Here is how a small pipeline with rule-based components and deep-learning components
    is distributed between the workers:

    <div style="text-align:center">
        <img src="/assets/images/model-parallelism.png" />
    </div>
    """
    executor = MultiprocessingStreamExecutor(stream)
    return executor.run()



import re
from typing import List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core.pipeline import PipelineProtocol
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.matchers.utils import Patterns
from edsnlp.pipes.base import BaseNERComponent
from edsnlp.utils.span_getters import SpanSetterArg

from .patterns import patterns


class SuicideAttemptMatcher(BaseNERComponent):
    """
    The `eds.suicide_attempt` pipeline component detects mentions of Suicide Attempt.
    It can be used with a span qualifier
    for contextualisation of the entity (history) and to detect false positives as
    negation, hypothesis or family. We recommend to use a machine learning qualifier
    to disambiguate polysemic words, as proposed in [@bey_natural_2024].


    Every matched entity will be labelled `suicide_attempt`.

    Extensions
    ----------
    Each entity span will have the suicide attempt modality as an attribute.
    The available modalities are:

    - `suicide_attempt_unspecific`
    - `autolysis`
    - `intentional_drug_overdose`
    - `jumping_from_height`
    - `cuts`
    - `strangling`
    - `self_destructive_behavior`
    - `burn_gas_caustic`

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.suicide_attempt())

    text = "J'ai vu le patient à cause d'une IMV avec paracetamol"
    doc = nlp(text)
    doc.ents
    # Out: (IMV,)

    ent = doc.ents[0]
    ent._.suicide_attempt_modality
    # Out: 'intentional_drug_overdose'
    ```

    ??? info "Patterns used for the named entity recognition"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/suicide_attempt/patterns.py"
        # fmt: on
        ```


    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name : str
        The name of the pipe
    attr : Union[str, Dict[str, str]]
        Attribute to match on, eg `TEXT`, `NORM`, etc.
    ignore_excluded : bool
        Whether to skip excluded tokens during matching.
    ignore_space_tokens : bool
        Whether to skip space tokens during matching.
    patterns : List[str]
        The regex pattern to use
    span_setter : SpanSetterArg
        How to set matches on the doc
    label : str
        Label name to use for the `Span` object and the extension

    Returns
    -------
    SuicideAttemptMatcher

    Authors and citation
    --------------------
    The `eds.suicide_attempt` component was developed by AP-HP's Data Science
    team, following the insights of the algorithm proposed
    by [@bey_natural_2024].
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        label: str = "suicide_attempt",
        name: Optional[str] = "eds.suicide_attempt",
        regex: Patterns = patterns,
        alignment_mode: str = "expand",
        attr: str = "TEXT",
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        flags: Union[re.RegexFlag, int] = 0,  # No additional flags
        span_from_group: bool = False,
        span_setter: SpanSetterArg = {"ents": True},
    ):
        self.attr = attr
        self.label = label
        self.regex_matcher = RegexMatcher(
            alignment_mode=alignment_mode,
            attr=attr,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
            flags=flags,
            span_from_group=span_from_group,
        )

        self.regex_matcher.build_patterns(regex=regex)

        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

    def set_extensions(self):
        if not Span.has_extension("suicide_attempt_modality"):
            Span.set_extension("suicide_attempt_modality", default=None)

    def process(self, doc: Doc) -> List[Span]:
        """
        Find matching spans in doc.

        Parameters
        ----------
        doc:
            Doc object.

        Returns
        -------
        spans:
            List of Spans returned by the matchers.
        """

        spans = list(self.regex_matcher(doc, as_spans=True))

        for span in spans:
            span._.suicide_attempt_modality = span.label_

            span.label_ = self.label

        return spans

    def __call__(self, doc: Doc) -> Doc:
        """
        Adds spans to document.

        Parameters
        ----------
        doc:
            Doc object

        Returns
        -------
        doc:
            Doc object, annotated with suicide attempts entities.
        """
        matches = self.process(doc)

        self.set_spans(doc, matches)

        return doc



import pytest
from confit import validate_arguments

import edsnlp
import edsnlp.pipes as eds
from edsnlp.utils.span_getters import (
    ContextWindow,
    get_spans,
    get_spans_with_group,
    make_span_context_getter,
    validate_span_setter,
)


def test_span_getter_dedupliation(lang):
    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.diabetes")

    doc = nlp("le patient a un diabète")

    span_getter = {"diabetes": True, "ents": True}

    assert len(list(get_spans(doc, span_getter, deduplicate=False))) == 2
    assert len(list(get_spans(doc, span_getter, deduplicate=True))) == 1


def test_span_context_getter(lang):
    nlp = edsnlp.blank(lang)
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.matcher", config={"terms": {"sentence": "sentence"}})
    doc = nlp(
        "This is a sentence. "
        "This is another sentence. "
        "This is a third one. "
        "Last sentence."
    )

    span_getter = make_span_context_getter(
        context_words=2,
        context_sents=1,
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This",
        "This is another sentence. This",
        ". Last sentence.",
    ]

    span_getter = make_span_context_getter(
        context_words=0,
        context_sents=1,
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence.",
        "This is another sentence.",
        "Last sentence.",
    ]

    span_getter = make_span_context_getter(
        context_words=0,
        context_sents=2,
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence.",
        "This is a sentence. This is another sentence. This is a third one.",
        "This is a third one. Last sentence.",
    ]


def test_span_getter_on_span():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.matcher(
            terms={"animal": ["snake", "dog"]},
            span_setter=["ents", "animals"],
        )
    )
    doc = nlp(
        "There was a snake. "
        "His friend was a dog. "
        "He liked baking cakes. "
        "But since he had no hands, he was a bad baker."
    )
    sents = list(doc.sents)

    # fmt: off
    assert str(list(get_spans(sents[0], validate_span_setter("ents")))) == "[snake]"
    assert str(list(get_spans(sents[0], validate_span_setter("animals")))) == "[snake]"
    assert str(list(get_spans(doc[5:], validate_span_setter("animals")))) == "[dog]"
    assert str(list(get_spans(doc[5:], validate_span_setter("*")))) == "[dog]"
    assert str(list(get_spans(doc[5:], validate_span_setter("sents")))) == "[There was a snake., His friend was a dog., He liked baking cakes., But since he had no hands, he was a bad baker.]"  # noqa: E501

    assert str(list(get_spans_with_group(sents[0], validate_span_setter("ents")))) == "[(snake, 'ents')]"  # noqa: E501
    assert str(list(get_spans_with_group(sents[0], validate_span_setter("animals")))) == "[(snake, 'animals')]"  # noqa: E501
    assert str(list(get_spans_with_group(doc[5:], validate_span_setter("animals")))) == "[(dog, 'animals')]"  # noqa: E501
    assert str(list(get_spans_with_group(doc[5:], validate_span_setter("*")))) == "[(dog, 'animals')]"  # noqa: E501
    assert str(list(get_spans_with_group(doc[5:], validate_span_setter("sents")))) == "[(There was a snake., 'sents'), (His friend was a dog., 'sents'), (He liked baking cakes., 'sents'), (But since he had no hands, he was a bad baker., 'sents')]"  # noqa: E501
    # fmt: on


def test_span_context_getter_asymmetric(lang):
    nlp = edsnlp.blank(lang)
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.matcher", config={"terms": {"animal": "kangaroo"}})
    doc = nlp(
        "This is a sentence. "
        "This is another sentence with a kangaroo. "
        "This is a third one. "
        "Last sentence."
    )

    span_getter = make_span_context_getter(context_words=2, context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "with a kangaroo. This",
    ]

    span_getter = make_span_context_getter(context_words=(2, 1), context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "with a kangaroo.",
    ]

    span_getter = make_span_context_getter(context_words=(1, 2), context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "a kangaroo. This",
    ]

    span_getter = make_span_context_getter(context_words=0, context_sents=(1, 2))
    assert [span_getter(s).text for s in doc.ents] == [
        "This is another sentence with a kangaroo. This is a third one.",
    ]

    span_getter = make_span_context_getter(context_words=0, context_sents=(2, 2))
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence with a kangaroo. This is a third one."  # noqa: E501
    ]

    span_getter = make_span_context_getter(context_words=0, context_sents=(1, 1))
    assert [span_getter(s).text for s in doc.ents] == [
        "This is another sentence with a kangaroo."
    ]

    span_getter = make_span_context_getter(context_words=(1000, 0), context_sents=0)
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence with a kangaroo"
    ]

    span_getter = make_span_context_getter(
        context_words=(1000, 0), context_sents=(1, 2)
    )
    assert [span_getter(s).text for s in doc.ents] == [
        "This is a sentence. This is another sentence with a kangaroo. This is a third one."  # noqa: E501
    ]


def test_context_getter_syntax():
    @validate_arguments
    def get_snippet(span, context: ContextWindow):
        return context(span)

    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.matcher", config={"terms": {"dog": "dog"}})
    doc = nlp(
        "There was a snake. "
        "His friend was a dog. "
        "He liked baking cakes. "
        "But since he had no hands, he was a bad baker. "
    )

    assert (
        get_snippet(doc.ents[0], "words[-5:5]").text
        == ". His friend was a dog. He liked baking cakes"
    )

    assert get_snippet(doc.ents[0], "words[-5:5] & sent").text == "His friend was a dog"

    assert (
        get_snippet(doc.ents[0], "words[-5:8] | sents[-1:1]").text
        == "There was a snake. His friend was a dog. He liked baking cakes. "
        "But since"
    )


def test_invalid_context_getter_syntax():
    @validate_arguments
    def apply_context(context: ContextWindow):
        pass

    apply_context("sents[-2:2]")

    with pytest.raises(ValueError):
        apply_context("stuff[-2:2]")



import ast
import importlib
import inspect
import logging
import sys
from typing import Any, Union

import astunparse
from griffe import (
    DocstringSectionParameters,
    Expr,
    Extension,
    Object,
    ObjectNode,
    patch_loggers,
)


def get_logger(name):
    new_logger = logging.getLogger(name)
    new_logger.setLevel("ERROR")
    return new_logger


patch_loggers(get_logger)

logger = get_logger(__name__)


class EDSNLPDocstrings(Extension):
    def __init__(self):
        super().__init__()

        self.PIPE_OBJ = {}
        self.FACT_MEM = {}
        self.PIPE_TO_FACT = {}

    def on_instance(
        self,
        node: Union[ast.AST, ObjectNode],
        obj: Object,
        agent: Any = None,
    ) -> None:
        if (
            isinstance(node, ast.Assign)
            and obj.name == "create_component"
            and isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Call)
        ):
            module_name = obj.path.rsplit(".", 1)[0]
            for name, mod in list(sys.modules.items()):
                if name.startswith("edspdf"):
                    importlib.reload(mod)
            module = importlib.reload(importlib.import_module(module_name))

            config_node = node.value.func
            config_node = next(
                (kw.value for kw in config_node.keywords if kw.arg == "default_config"),
                None,
            )
            try:
                default_config = eval(astunparse.unparse(config_node), module.__dict__)
            except Exception:
                default_config = {}

            # import object to get its evaluated docstring
            try:
                runtime_obj = getattr(module, obj.name)
                source = inspect.getsource(runtime_obj)
                self.visit(ast.parse(source))
            except ImportError:
                logger.debug(f"Could not get dynamic docstring for {obj.path}")
                return
            except AttributeError:
                logger.debug(f"Object {obj.path} does not have a __doc__ attribute")
                return

            callee = (
                runtime_obj.__init__
                if hasattr(runtime_obj, "__init__")
                else runtime_obj
            )
            spec = inspect.getfullargspec(callee)
            func_defaults = dict(
                zip(spec.args[-len(callee.__defaults__) :], callee.__defaults__)
                if callee.__defaults__
                else (),
                **(callee.__kwdefaults__ or {}),
            )
            defaults = {**func_defaults, **default_config}
            self.FACT_MEM[obj.path] = (node, obj, defaults)
            pipe_path = runtime_obj.__module__ + "." + runtime_obj.__name__
            self.PIPE_TO_FACT[pipe_path] = obj.path

            if pipe_path in self.PIPE_OBJ:
                pipe = self.PIPE_OBJ[pipe_path]
                obj.docstring = pipe.docstring
            else:
                return
        elif obj.is_class or obj.is_function:
            self.PIPE_OBJ[obj.path] = obj
            if obj.path in self.PIPE_TO_FACT:
                node, fact_obj, defaults = self.FACT_MEM[self.PIPE_TO_FACT[obj.path]]
                fact_obj.docstring = obj.docstring
                obj = fact_obj
            else:
                return
        else:
            return

        if obj.docstring is None:
            return

        param_section: DocstringSectionParameters = None
        obj.docstring.parser = "numpy"
        for section in obj.docstring.parsed:
            if isinstance(section, DocstringSectionParameters):
                param_section = section  # type: ignore

        if param_section is None:
            return

        for param in param_section.value:
            if param.name in defaults:
                param.default = str(defaults[param.name])
            if isinstance(param.default, Expr):
                continue
            if param.default is not None and len(param.default) > 50:
                param.default = param.default[: 50 - 3] + "..."



from edsnlp import registry

from .span_classifier import TrainableSpanClassifier

create_component = registry.factory.register(
    "eds.span_classifier",
    assigns=[],
    deprecated=["eds.span_qualifier"],
)(TrainableSpanClassifier)



from pytest import mark

import edsnlp

text = (
    "Le patient est admis pour des douleurs dans le bras droit. "
    "mais n'a pas de problème de locomotion. \n"
    "Historique d'AVC dans la famille\n"
    "Mais ne semble pas en être un\n"
    "Pourrait être un cas de rhume.\n"
    "Motif :\n"
    "Douleurs dans le bras droit !"
    "Il est contaminé à E.Coli? c'est un problème, il faut s'en occuper."
)


@mark.parametrize("endlines", [True, False])
def test_sentences(endlines):
    nlp = edsnlp.blank("fr")
    nlp.add_pipe("eds.sentences", config={"use_endlines": endlines})
    doc = nlp.make_doc(text)

    if endlines:
        doc[28].tag_ = "EXCLUDED"

    doc = nlp(doc)

    sents_text = [sent.text for sent in doc.sents]

    if endlines:
        assert sents_text == [
            "Le patient est admis pour des douleurs dans le bras droit.",
            "mais n'a pas de problème de locomotion. \n",
            "Historique d'AVC dans la famille\nMais ne semble pas en être un\n",
            "Pourrait être un cas de rhume.\n",
            "Motif :\n",
            "Douleurs dans le bras droit !",
            "Il est contaminé à E.Coli?",
            "c'est un problème, il faut s'en occuper.",
        ]
    else:
        assert sents_text == [
            "Le patient est admis pour des douleurs dans le bras droit.",
            "mais n'a pas de problème de locomotion. \n",
            "Historique d'AVC dans la famille\n",
            "Mais ne semble pas en être un\n",
            "Pourrait être un cas de rhume.\n",
            "Motif :\n",
            "Douleurs dans le bras droit !",
            "Il est contaminé à E.Coli?",
            "c'est un problème, il faut s'en occuper.",
        ]

    nlp("")


def test_false_positives(blank_nlp):
    false_positives = [
        "02.04.2018",
        "E.Coli",
    ]

    for fp in false_positives:
        doc = blank_nlp(fp)
        assert len(list(doc.sents)) == 1


def test_newlines_double():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        "eds.sentences",
        config={
            "punct_chars": [],
            "ignore_excluded": False,
            "check_capitalized": False,
            "min_newline_count": 2,
        },
    )

    doc = nlp(
        """\
Lundi
Mardi
Mercredi
Le patient est admis. Des douleurs dans le bras droit
\n\n
jeudi."""
    )
    assert len(list(doc.sents)) == 2

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        "eds.sentences",
        config={
            "punct_chars": [],
            "ignore_excluded": False,
            "check_capitalized": True,
            "min_newline_count": 2,
        },
    )

    doc = nlp(
        """\
Lundi
Mardi
Mercredi
Le patient est admis. Des douleurs dans le bras droit
\n
jeudi."""
    )
    assert len(list(doc.sents)) == 1



import json
import math
import os
import time
import warnings
from collections import defaultdict
from contextlib import nullcontext
from itertools import chain
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Union,
)

import torch
from accelerate import Accelerator, PartialState
from accelerate.tracking import GeneralTracker
from accelerate.utils import gather_object
from confit import Draft, validate_arguments
from confit.utils.random import set_seed
from tqdm import tqdm, trange
from typing_extensions import Literal

import edsnlp
from edsnlp import Pipeline, registry
from edsnlp.core.stream import Stream
from edsnlp.metrics.ner import NerMetric
from edsnlp.metrics.span_attribute import SpanAttributeMetric
from edsnlp.pipes.base import (
    BaseNERComponent,
    BaseSpanAttributeClassifierComponent,
)
from edsnlp.utils.batching import BatchSizeArg, stat_batchify
from edsnlp.utils.bindings import BINDING_SETTERS
from edsnlp.utils.collections import (
    chain_zip,
    decompress_dict,
    flatten,
    flatten_once,
    ld_to_dl,
)
from edsnlp.utils.span_getters import get_spans
from edsnlp.utils.typing import AsList

from ..core.torch_component import TorchComponent
from .optimizer import LinearSchedule, ScheduledOptimizer


def deep_add_flat(x, result, path=(), check_stats_key=True):
    if result is None:
        result = {}
    if isinstance(x, dict):
        for k, v in x.items():
            deep_add_flat(v, result, (*path, k), check_stats_key=check_stats_key)
        return result
    if not check_stats_key or "stats" in path and "__batch_hash__" not in path[-1]:
        if isinstance(x, (float, int)):
            path = "/".join(path)
            result[path] = result.get(path, 0) + x
        elif isinstance(x, torch.Tensor) and x.ndim == 0:
            path = "/".join(path)
            result[path] = result.get(path, 0) + x.item()
    return result


def deep_set_unflat(x, stats):
    for k, v in stats.items():
        path = k.split("/")
        current = x
        for p in path[:-1]:
            if p not in current:
                break
            current = current[p]
        else:
            current[path[-1]] = v
    return x


@validate_arguments
class GenericScorer:
    def __init__(
        self,
        batch_size: Union[int, str] = 1,
        autocast: Union[bool, Any] = None,
        speed: bool = True,
        **metrics,
    ):
        """
        A scorer to evaluate the model performance on various tasks.

        Parameters
        ----------
        batch_size: Union[int, str]
            The batch size to use for scoring. Can be an int (number of documents)
            or a string (batching expression like "2000 words").
        speed: bool
            Whether to compute the model speed (words/documents per second)
        autocast: Union[bool, Any]
            Whether to use autocasting for mixed precision during the evaluation,
            defaults to True.
        metrics: Dict[str, Any]
            A keyword arguments mapping of metric names to metrics objects. See the
            [metrics](/metrics) documentation for more info.
        """
        self.metrics = metrics
        self.speed = speed
        self.batch_size = batch_size
        self.autocast = autocast

    def __call__(self, nlp: Pipeline, docs: Iterable[Any]):
        scores = {}
        docs = list(docs)
        metrics = dict(self.metrics)

        # Speed
        if self.speed:
            t0 = time.time()
            list(
                nlp.pipe(
                    d.copy() for d in tqdm(docs, desc="Computing model speed")
                ).set_processing(
                    batch_size=self.batch_size,
                    autocast=self.autocast,
                )
            )
            duration = time.time() - t0
            scores["speed"] = dict(
                wps=sum(len(d) for d in docs) / duration,
                dps=len(docs) / duration,
            )

        # NER
        ner_pipes = [
            name for name, pipe in nlp.pipeline if isinstance(pipe, BaseNERComponent)
        ]
        ner_metrics = {
            name: metrics.pop(name)
            for name in list(metrics)
            if isinstance(metrics[name], NerMetric)
        }
        if ner_pipes and ner_metrics:
            clean_ner_docs = [d.copy() for d in tqdm(docs, desc="Copying docs")]
            for d in clean_ner_docs:
                d.ents = []
                d.spans.clear()
            with nlp.select_pipes(enable=ner_pipes):
                ner_preds = list(
                    nlp.pipe(tqdm(clean_ner_docs, desc="Predicting")).set_processing(
                        batch_size=self.batch_size,
                        autocast=self.autocast,
                    )
                )
            for name, metric in ner_metrics.items():
                scores[name] = metric(docs, ner_preds)

        # Qualification
        qlf_pipes = [
            name
            for name, pipe in nlp.pipeline
            if isinstance(pipe, BaseSpanAttributeClassifierComponent)
        ]
        span_attr_metrics = {
            name: metrics.pop(name)
            for name in list(metrics)
            if isinstance(metrics[name], SpanAttributeMetric)
        }
        if qlf_pipes and span_attr_metrics:
            clean_qlf_docs = [d.copy() for d in tqdm(docs, desc="Copying docs")]
            for doc in clean_qlf_docs:
                for name in qlf_pipes:
                    pipe = nlp.get_pipe(name)
                    for span in get_spans(doc, pipe.span_getter):
                        for qlf in nlp.get_pipe(name).attributes:
                            BINDING_SETTERS[(qlf, None)](span)
            with nlp.select_pipes(disable=ner_pipes):
                qlf_preds = list(
                    nlp.pipe(tqdm(clean_qlf_docs, desc="Predicting")).set_processing(
                        batch_size=self.batch_size,
                        autocast=self.autocast,
                    )
                )
            for name, metric in span_attr_metrics.items():
                scores[name] = metric(docs, qlf_preds)

        # Custom metrics
        for name, metric in metrics.items():
            pred_docs = [d.copy() for d in tqdm(docs, desc="Copying docs")]
            preds = list(
                nlp.pipe(tqdm(pred_docs, desc="Predicting")).set_processing(
                    batch_size=self.batch_size,
                    autocast=self.autocast,
                )
            )
            scores[name] = metric(docs, preds)

        return scores


if TYPE_CHECKING:
    GenericScorer = Union[GenericScorer, Dict]


def ewm_moments(x, window, adjust=True, bias=False, state=None):
    if state is None:
        alpha = 2.0 / (window + 1)
        decay = 1 - alpha
        fresh_weight = 1 if adjust else alpha
        mean_val = x
        var_val = 0.0
        sum_w = 1.0
        sum_w2 = 1.0
        old_w = 1.0
        return (
            mean_val,
            float("nan"),
            [decay, fresh_weight, mean_val, var_val, sum_w, sum_w2, old_w],
        )
    else:
        decay, fresh_weight, mean_val, var_val, sum_w, sum_w2, old_w = state

    sum_w *= decay
    sum_w2 *= decay * decay
    old_w *= decay
    old_m = mean_val
    denom = old_w + fresh_weight
    mean_val = (old_w * old_m + fresh_weight * x) / denom
    d1 = old_m - mean_val
    d2 = x - mean_val
    var_val = (old_w * (var_val + d1 * d1) + fresh_weight * d2 * d2) / denom
    sum_w += fresh_weight
    sum_w2 += fresh_weight * fresh_weight
    old_w += fresh_weight

    state = [decay, fresh_weight, mean_val, var_val, sum_w, sum_w2, old_w]

    if not bias:
        num = sum_w * sum_w
        den = num - sum_w2
        var_val = var_val * (num / den) if den > 0 else float("nan")

    return mean_val, var_val, state


def default_optim(
    trained_pipes,
    *,
    task_lr: float = 3e-4,
    transformer_lr: float = 5e-5,
    warmup_rate: float = 0.1,
    max_steps: int,
):
    from edsnlp.pipes.trainable.embeddings.transformer.transformer import Transformer

    trf_pipe = next(
        (
            module
            for pipe in trained_pipes
            for name, module in pipe.named_component_modules()
            if isinstance(module, Transformer)
        ),
        None,
    )
    params = set(p for pipe in trained_pipes for p in pipe.parameters())
    trf_params = params & set(trf_pipe.parameters() if trf_pipe else ())

    return ScheduledOptimizer(
        torch.optim.AdamW(
            [
                {
                    "params": list(params - trf_params),
                    "lr": task_lr,
                    "schedules": LinearSchedule(
                        total_steps=max_steps,
                        warmup_rate=warmup_rate,
                        start_value=task_lr,
                    ),
                }
            ]
            + [
                {
                    "params": list(trf_params),
                    "lr": transformer_lr,
                    "schedules": LinearSchedule(
                        total_steps=max_steps,
                        warmup_rate=warmup_rate,
                        start_value=0,
                    ),
                },
            ][: 1 if transformer_lr else 0]
        )
    )


@validate_arguments
class TrainingData:
    def __init__(
        self,
        data: Stream,
        batch_size: BatchSizeArg,
        shuffle: Union[str, Literal[False]],
        sub_batch_size: Optional[BatchSizeArg] = None,
        pipe_names: Optional[AsList[str]] = None,
        post_init: bool = True,
    ):
        """
        A training data object.

        Parameters
        ----------
        data: Stream
            The stream of documents to train on. The documents will be
            preprocessed and collated according to the pipeline's components.
        batch_size: BatchSizeArg
            The batch size. Can be a batching expression like "2000 words",
            an int (number of documents), or a tuple (batch_size, batch_by).
            The batch_by argument should be a statistic produced by the
            pipes that will be trained. For instance, the `eds.span_pooler`
            component produces a "spans" statistic, that can be used to
            produce batches of no more than 16 spans by setting batch_size
            to "16 spans".
        shuffle: str
            The shuffle strategy. Can be "dataset" to shuffle the entire
            dataset (this can be memory-intensive for large file based
            datasets), "fragment" to shuffle the fragment-based datasets
            like parquet files, or a batching expression like "2000 words"
            to shuffle the dataset in chunks of 2000 words.
        sub_batch_size: Optional[BatchSizeArg]
            How to split each batch into sub-batches that will be fed to
            the model independently to accumulate gradients over.
            To split a batch of 8000 tokens into smaller batches of 1000
            tokens each, just set this to "1000 tokens".

            You can also request a number of splits, like "4 splits",
            to split the batch into N parts each close to (but less than)
            batch_size / N.
        pipe_names: Optional[Collection[str]]
            The names of the pipes that should be trained on this data.
            If None, defaults to all trainable pipes.
        post_init: bool
            Whether to call the pipeline's post_init method with the data
            before training.
        """
        self.data = data
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.sub_batch_size = sub_batch_size
        self.pipe_names = set(pipe_names) if pipe_names else None
        self.post_init = post_init

    def __call__(self, nlp, device):
        data = self.data.loop()
        if self.shuffle:
            data = data.shuffle(self.shuffle)

        with nlp.select_pipes(enable=self.pipe_names):
            data = data.map(nlp.preprocess, kwargs=dict(supervision=True))
        batcher = stat_batchify(self.batch_size[1] or "docs")
        if self.sub_batch_size and self.sub_batch_size[1] == "splits":
            data = data.batchify(
                batch_size=self.batch_size[0] // self.sub_batch_size[0],
                batch_by=batcher,
            )
            data = data.batchify(batch_size=self.sub_batch_size[0])
            data = data.map(lambda b: [nlp.collate(sb, device=device) for sb in b])
        elif self.sub_batch_size:
            data = data.batchify(batch_size=self.batch_size[0], batch_by=batcher)
            sub_batcher = stat_batchify(self.sub_batch_size[1] or "docs")
            data = data.map(
                lambda batch: [
                    nlp.collate(sub_batch, device=device)
                    for sub_batch in sub_batcher(batch, self.sub_batch_size[0])
                ]
            )
        else:
            data = data.batchify(batch_size=self.batch_size[0], batch_by=batcher)
            data = data.map(nlp.collate, kwargs=dict(device=device))
        return data


class PipeDict(torch.nn.ModuleDict):
    def __init__(self, pipes, loss_scales):
        super().__init__(pipes)
        self.loss_scales = loss_scales

    def forward(self, batch, enable: Optional[Sequence[str]] = None):
        loss = None
        all_results = {}
        for name, pipe in self.items():
            if enable is None or name in enable:
                res = pipe(batch[name])
                all_results[name] = res
                if "loss" in res:
                    res["loss"] = res["loss"] * self.loss_scales.get(name, 1)
                    loss = res["loss"] if loss is None else loss + res["loss"]
                    if torch.isnan(loss):
                        raise ValueError(f"NaN loss at component {name}")
        return all_results, loss


def get_logger(
    logger: Union[bool, AsList[Union[str, Draft[GeneralTracker], GeneralTracker]]],
    project_name,
    logging_dir,
    **kwargs,
) -> List[GeneralTracker]:
    logger = ["rich", "json"] if logger is True else [] if not logger else logger
    logger_drafts_maybe = []
    for log in logger:
        if isinstance(log, str):
            cls = edsnlp.registry.loggers.get(log)
            logger_drafts_maybe.append(Draft(cls, {}))
        else:
            logger_drafts_maybe.append(log)
    logger = logger_drafts_maybe
    logger = [
        Draft.instantiate(obj, project_name=project_name, logging_dir=logging_dir)
        for obj in logger
    ]
    return logger


@validate_arguments(registry=registry)
def train(
    *,
    nlp: Pipeline,
    train_data: AsList[TrainingData],
    val_data: AsList[Stream] = [],
    seed: int = 42,
    max_steps: int = 1000,
    optimizer: Union[
        Draft[ScheduledOptimizer],
        ScheduledOptimizer,
        torch.optim.Optimizer,
    ] = None,
    validation_interval: Optional[int] = None,
    checkpoint_interval: Optional[int] = None,
    grad_max_norm: float = 5.0,
    grad_ewm_window: int = 100,
    grad_dev_policy: Optional[Literal["clip_mean", "clip_threshold", "skip"]] = None,
    grad_max_dev: float = 7.0,
    loss_scales: Dict[str, float] = {},
    scorer: GenericScorer = GenericScorer(),
    num_workers: int = 0,
    cpu: bool = False,
    mixed_precision: Literal["no", "fp16", "bf16", "fp8"] = "no",
    output_dir: Union[Path, str] = Path("artifacts"),
    output_model_dir: Optional[Union[Path, str]] = None,
    save_model: bool = True,
    logger: Union[bool, AsList[Union[str, GeneralTracker, Draft[GeneralTracker]]]] = True,  # noqa: E501
    log_weight_grads: bool = False,
    on_validation_callback: Optional[Callable[[Dict], None]] = None,
    project_name: str = None,
    config_meta: Optional[Dict] = None,
    **kwargs,
):  # fmt: skip
    """
    Train a pipeline.

    Parameters
    ----------
    nlp: Pipeline
        The pipeline that will be trained in place.
    train_data: AsList[TrainingData]
        The training data. Can be a single
        [TrainingData][edsnlp.training.trainer.TrainingData] object, a dict that
        will be cast or a list of these objects.

        ??? note "`TrainingData` object/dictionary"
            ::: edsnlp.training.trainer.TrainingData
                options:
                    heading_level: 1
                    only_parameters: "no-header"
                    skip_parameters: []
                    show_source: false
                    show_toc: false
    val_data: AsList[Stream]
        The validation data. Can be a single Stream object or a list of
        Stream.
    seed: int
        The random seed
    max_steps: int
        The maximum number of training steps
    optimizer: Union[ScheduledOptimizer, Draft[ScheduledOptimizer], torch.optim.Optimizer]
        The optimizer. If None, a default optimizer will be used.

        ??? note "`ScheduledOptimizer` object/dictionary"
            ::: edsnlp.training.optimizer.ScheduledOptimizer
                options:
                    heading_level: 1
                    only_parameters: "no-header"
                    skip_parameters: []
                    show_source: false
                    show_toc: false
    validation_interval: Optional[int]
        The number of steps between each evaluation. Defaults to 1/10 of max_steps
    checkpoint_interval: Optional[int]
        The number of steps between each model save. Defaults to validation_interval
    grad_max_norm: float
        The maximum gradient norm
    grad_dev_policy: Optional[Literal["clip_mean", "clip_threshold"]]
        The policy to apply when a gradient spike is detected, ie. when the
        gradient norm is higher than the mean + std * grad_max_dev. Can be:

        - "clip_mean": clip the gradients to the mean gradient norm
        - "clip_threshold": clip the gradients to the mean + std * grad_max_dev
        - "skip": skip the step

        These do not apply to `grad_max_norm` that is always enforced when it is not
        None, since `grad_max_norm` is not adaptive and would most likely prohibit
        the model from learning during the early stages of training when gradients are
        expected to be high.
    grad_ewm_window: int
        Approximately how many steps should we look back to compute the average
        gradient norm and variance to detect gradient deviation spikes.
    grad_max_dev: float
        The threshold to apply to detect gradient spikes. A spike is detected
        when the value is higher than the mean + variance * threshold.
    loss_scales: Dict[str, float]
        The loss scales for each component (useful for multi-task learning)
    scorer: GenericScorer
        How to score the model. Expects a `GenericScorer` object or a dict
        containing a mapping of metric names to metric objects.

        ??? note "`GenericScorer` object/dictionary"
            ::: edsnlp.training.trainer.GenericScorer
                options:
                    heading_level: 1
                    only_parameters: "no-header"
                    skip_parameters: []
                    show_source: false
                    show_toc: false
    num_workers: int
        The number of workers to use for preprocessing the data in parallel.
        Setting it to 0 means no parallelization : data is processed on the
        main thread which may induce latency slow down the training. To
        avoid this, a good practice consist in doing the preprocessing either
        before training or in parallel in a separate process. Because of how
        EDS-NLP handles stream multiprocessing, changing this value
        will affect the order of the documents in the produces batches.
        A stream [1, 2, 3, 4, 5, 6] split in batches of size 3 will produce:

        - [1, 2, 3] and [4, 5, 6] with 1 worker
        - [1, 3, 5] and [2, 4, 6] with 2 workers
    cpu: bool
        Whether to use force training on CPU. On MacOS, this might be
        necessary to get around some `mps` backend issues.
    mixed_precision: Literal["no", "fp16", "bf16", "fp8"]
        The mixed precision mode. Can be "no", "fp16", "bf16" or "fp8".
    output_dir: Union[Path, str]
        The output directory, which will contain a `model-last` directory
        with the last model, and a `train_metrics.json` file with the
        training metrics and stats.
    output_model_dir: Optional[Union[Path, str]]
        The directory where to save the model. If None, defaults to
        `output_dir / "model-last"`.
    save_model: bool
        Whether to save the model or not. This can be useful if you are only
        interested in the metrics, but no the model, and want to avoid
        spending time dumping the model weights to the disk.
    logger: Union[bool, AsList[Union[str, Partial[GeneralTracker], GeneralTracker]]]
        The logger to use. Can be a boolean to use the default loggers (rich
        and json), a list of logger names, or a list of logger objects.

        You can use huggingface accelerate integrated loggers (`tensorboard`,
        `wandb`, `comet_ml`, `aim`, `mlflow`, `clearml`, `dvclive`), or
        EDS-NLP simple loggers, or a combination of both:

        - `csv`: logs to a CSV file in `output_dir` (`artifacts/metrics.csv`)
        - `json`: logs to a JSON file in `output_dir` (`artifacts/metrics.json`)
        - `rich`: logs to a rich table in the terminal
    log_weight_grads: bool
        Whether to log the weight gradients during training.
    on_validation_callback: Optional[Callable[[Dict], None]]
        A callback function invoked during validation steps to handle custom logic.
    project_name: str
        The project name, used to group experiments in some loggers. If None,
        defaults to the path of the config file, relative to the home directory, with
        slashes replaced by double underscores.
    kwargs: Dict
        Additional keyword arguments.

    Returns
    -------
    Pipeline
        The trained pipeline
    """  # noqa: E501
    # hack to ensure cpu is set before the accelerator is indirectly initialized
    # when creating the trackers
    PartialState(cpu=cpu)
    project_name = project_name or str(
        os.curdir if config_meta is None else config_meta["config_path"][0]
    ).replace("/", "__")
    output_dir = Path(output_dir or Path.cwd() / "artifacts")
    output_model_dir = Path(output_model_dir or output_dir / "model-last")
    accelerator = Accelerator(
        cpu=cpu,
        mixed_precision=mixed_precision,
        log_with=get_logger(
            logger,
            # default project name, the user can override this when creating the logger
            project_name=project_name,
            logging_dir=output_dir,
        ),
    )
    # accelerator.register_for_checkpointing(dataset)
    is_main_process = accelerator.is_main_process
    device = accelerator.device

    if "max_grad_norm" in kwargs:
        warnings.warn(
            "The 'max_grad_norm' argument is deprecated. Use 'grad_max_norm' instead."
        )
        grad_max_norm = kwargs.pop("max_grad_norm")

    unresolved_config = None
    if is_main_process:
        os.makedirs(output_dir, exist_ok=True)
        os.makedirs(output_model_dir, exist_ok=True)
        if config_meta is not None:  # pragma: no cover
            unresolved_config = config_meta["unresolved_config"]
            unresolved_config["train"]["project_name"] = project_name
            print(unresolved_config.to_yaml_str())
            unresolved_config.to_disk(output_dir / "train_config.yml")
        # TODO: handle config_meta is None
    accelerator.init_trackers(
        project_name,
        config=json.loads(json.dumps(unresolved_config, default=str)),
    )  # in theory project name shouldn't be used

    validation_interval = validation_interval or max_steps // 10
    checkpoint_interval = checkpoint_interval or validation_interval
    trainable_pipe_names = {name for name, pipe in nlp.torch_components()}
    phases = nlp.connected_pipes_names()
    accelerator.print("Trainable components: " + ", ".join(trainable_pipe_names))
    accelerator.print(
        "Training phases:"
        + "".join(f"\n - {i + 1}: {', '.join(n)}" for i, n in enumerate(phases))
    )

    if kwargs:
        raise ValueError(f"Unknown arguments: {', '.join(kwargs)}")

    # Prepare validation docs
    val_docs = list(chain.from_iterable(val_data))

    # Initialize pipeline with training documents
    nlp.post_init(chain_zip([td.data for td in train_data if td.post_init]))

    all_params = set(nlp.parameters())
    optim = optimizer
    del optimizer
    optim = Draft.instantiate(optim, module=nlp, total_steps=max_steps)
    if optim is None:
        warnings.warn(
            "No optimizer provided, using default optimizer with default parameters"
        )
        optim = default_optim(
            [nlp.get_pipe(name) for name in trainable_pipe_names],
            max_steps=max_steps,
            **{
                k: kwargs.pop(k)
                for k in ("task_lr", "transformer_lr", "warmup_rate")
                if k in kwargs
            },
        )
    optim: torch.nn.Optimizer = Draft.instantiate(
        optim,
        module=nlp,
        total_steps=max_steps,
    )

    for td in train_data:
        if not (td.pipe_names is None or td.pipe_names <= trainable_pipe_names):
            raise ValueError(
                f"Training data pipe names {td.pipe_names} should be a subset of "
                f"the trainable pipe names {trainable_pipe_names}, or left to None "
                f"use this dataset for all trainable components."
            )

    for phase_i, pipe_names in enumerate(phases):
        trained_pipes_local: Dict[str, TorchComponent] = {
            n: nlp.get_pipe(n) for n in pipe_names
        }
        trained_pipes = PipeDict(trained_pipes_local, loss_scales)
        trained_pipes_params = set(trained_pipes.parameters())
        phase_training_data = [
            td
            for td in train_data
            if td.pipe_names is None or set(td.pipe_names) & set(pipe_names)
        ]

        if len(phase_training_data) == 0:
            raise ValueError(
                f"No training data found for phase {phase_i + 1} with components "
                f"{', '.join(pipe_names)}. Make sure that these components are "
                f"listed in the 'pipe_names' attribute of at least one of the "
                f"provided training data."
            )

        with nlp.select_pipes(disable=trainable_pipe_names - set(pipe_names)):
            accelerator.print(f"Phase {phase_i + 1}: training {', '.join(pipe_names)}")
            set_seed(seed)

            optim_params = {p for g in optim.param_groups for p in g["params"]}
            grad_params = set()
            for param in all_params:
                has_grad_param = param in optim_params and param in trained_pipes_params
                if has_grad_param:
                    grad_params.add(param)
                param.requires_grad_(has_grad_param)

            accelerator.print("Optimizing groups:")
            for g in optim.param_groups:
                accelerator.print(
                    " - {} weight tensors ({:,} parameters){}".format(
                        len([p for p in g["params"] if p in grad_params]),
                        sum([p.numel() for p in g["params"] if p in grad_params]),
                        ": " + " & ".join(g.get("selectors", "*"))
                        if "selectors" in g
                        else "",
                    )
                )
            accelerator.print(
                ("! WARNING ! " if (len(all_params - grad_params) > 0) else "")
                + f"Keeping frozen {len(all_params - grad_params):} weight tensors "
                f"({sum(p.numel() for p in all_params - grad_params):,} parameters)"
            )

            nlp.train(True)

            phase_datasets = [
                td(nlp, device).set_processing(
                    num_cpu_workers=num_workers,
                    process_start_method="spawn",
                )
                for td in phase_training_data
            ]
            iterator = iter(zip(*(phase_datasets)))
            (accel_optim, trained_pipes) = accelerator.prepare(optim, trained_pipes)
            if hasattr(accel_optim.optimizer, "initialize"):
                accel_optim.optimizer.initialize()

            ewm_state = grad_mean = grad_var = None
            count = 0
            spikes = 0
            flat_cum_stats = defaultdict(float)
            flat_cum_res = defaultdict(float)
            flat_cum_weight_metrics = defaultdict(float)
            set_seed(seed)
            # Training loop
            for step in trange(
                max_steps + 1,
                desc="Training model",
                leave=True,
                mininterval=5.0,
                total=max_steps,
                disable=not is_main_process,
                smoothing=0.3,
            ):
                if save_model and is_main_process and (step % checkpoint_interval) == 0:
                    nlp.to_disk(output_model_dir)
                if step > 0 and (step % validation_interval) == 0:
                    flat_cum_res = {
                        k: sum(v)
                        for k, v in ld_to_dl(gather_object([flat_cum_res])).items()
                    }
                    cum_res = decompress_dict(flat_cum_res)
                    if is_main_process:
                        cum_stats = decompress_dict(flat_cum_stats)
                        val_metrics = scorer(nlp, val_docs) if val_docs else {}
                        metrics = {
                            "step": step,
                            "lr": accel_optim.param_groups[0]["lr"],
                            "loss": cum_res["__all__"]["loss"] / count,
                            "count": count,
                            "spikes": spikes,
                            "weights": {
                                k: v / count for k, v in flat_cum_weight_metrics.items()
                            },
                            "per_pipe": {
                                n: {
                                    "stats": cum_stats[n].get("stats", {}),
                                    "results": (
                                        {
                                            k: v
                                            for k, v in trained_pipes_local[n]
                                            .compute_training_metrics(
                                                cum_res[n],
                                                cum_stats[n].get("stats", {}),
                                                count,
                                            )
                                            .items()
                                            if v is not None
                                        }
                                    ),
                                }
                                for n in pipe_names
                            },
                            "validation": val_metrics,
                        }
                        accelerator.log(metrics, step=step)
                    count = 0
                    spikes = 0
                    flat_cum_stats = defaultdict(float)
                    flat_cum_res = defaultdict(float)
                    flat_cum_weight_metrics = defaultdict(float)

                    if on_validation_callback:
                        on_validation_callback(metrics)

                if step == max_steps:
                    break

                accel_optim.zero_grad()

                batches = list(next(iterator))
                batches_pipe_names = list(
                    flatten_once(
                        [
                            [td.pipe_names or pipe_names] * len(b)
                            for td, b in zip(phase_training_data, batches)
                        ]
                    )
                )
                batches = list(flatten(batches))

                # Synchronize stats between sub-batches across workers
                local_batch_stats = {}
                for b in batches:
                    deep_add_flat(b, result=local_batch_stats)
                flat_batch_stats = ld_to_dl(gather_object([local_batch_stats]))
                flat_batch_stats = {k: sum(v) for k, v in flat_batch_stats.items()}

                # for training purposes (ie have the right denominator in the losses)
                for b in batches:
                    deep_set_unflat(b, flat_batch_stats)

                # for logging purposes
                for k, v in flat_batch_stats.items():
                    flat_cum_stats[k] += flat_batch_stats[k]

                for idx, (batch, batch_pipe_names) in enumerate(
                    zip(batches, batches_pipe_names)
                ):
                    cache_ctx = (
                        nlp.cache() if len(batch_pipe_names) > 1 else nullcontext()
                    )
                    no_sync_ctx = (
                        accelerator.no_sync(trained_pipes)
                        if idx < len(batches) - 1
                        else nullcontext()
                    )
                    try:
                        with cache_ctx, no_sync_ctx:
                            all_res, loss = trained_pipes(
                                batch,
                                enable=batch_pipe_names,
                            )
                            deep_add_flat(all_res, flat_cum_res, check_stats_key=False)
                            flat_cum_res["__all__/loss"] += loss.item()
                            del all_res
                            if isinstance(loss, torch.Tensor) and loss.requires_grad:
                                # Trick to ensure all trained parameters participate in
                                # the loss otherwise torch (in parallel) isn't happy as
                                # it cannot reliably know if a param gradients haven't
                                # been collected yet, or won't participate in the step
                                # at all.
                                loss0 = sum(p.sum() * 0 for p in grad_params)
                                accelerator.backward(loss + loss0)
                    except torch.cuda.OutOfMemoryError:  # pragma: no cover
                        print(
                            "Out of memory error encountered when processing a "
                            "batch with the following statistics:"
                        )
                        print(local_batch_stats)
                        raise
                    except Exception:
                        print(
                            "An error occurred when processing a batch with these"
                            "dimensions"
                        )
                        print(local_batch_stats)
                        raise

                    del loss

                del flat_batch_stats
                accelerator.unscale_gradients()

                # Log gradients
                if log_weight_grads and is_main_process:
                    for pipe_name, pipe in trained_pipes_local.items():
                        for param_name, param in pipe.named_parameters():
                            if param.grad is not None:
                                flat_cum_weight_metrics[
                                    f"grad_norm/{pipe_name}/{param_name}"
                                ] += param.grad.norm().item()
                                flat_cum_weight_metrics[
                                    f"param_norm/{pipe_name}/{param_name}"
                                ] += param.norm().item()

                grad_norm = torch.nn.utils.clip_grad_norm_(
                    grad_params, grad_max_norm, norm_type=2
                ).item()

                # Detect grad spikes and skip the step if necessary
                if grad_dev_policy is not None:
                    if step > grad_ewm_window and (
                        grad_norm - grad_mean
                    ) > grad_max_dev * math.sqrt(grad_var):
                        spike = True
                        spikes += 1
                    else:
                        grad_mean, grad_var, ewm_state = ewm_moments(
                            grad_norm, grad_ewm_window, state=ewm_state
                        )
                        spike = False

                    if spike and grad_dev_policy == "clip_mean":
                        torch.nn.utils.clip_grad_norm_(
                            grad_params, grad_mean, norm_type=2
                        )
                    elif spike and grad_dev_policy == "clip_threshold":
                        torch.nn.utils.clip_grad_norm_(
                            grad_params,
                            grad_mean + math.sqrt(grad_var) * grad_max_dev,
                            norm_type=2,
                        )

                if grad_dev_policy != "skip" or not spike:
                    accel_optim.step()

                count += 1
                flat_cum_weight_metrics["grad_norm/__all__"] += grad_norm

            del iterator

    # Should we put this in a finally block?
    accelerator.end_training()

    return nlp



"""
Metrics for Span Attribute Classification

# Span Attribute Classification Metrics {: #edsnlp.metrics.span_attribute.SpanAttributeMetric }

Several NLP tasks consist in classifying existing spans of text into multiple classes,
such as the detection of negation, hypothesis or span linking.

We provide a metric to evaluate the performance of such tasks,

Let's look at an example:

+-------------------------------------------------------------------+-------------------------------------------------------------------+
| pred                                                              | ref                                                               |
+===================================================================+===================================================================+
| Le patient n'est pas *fièvreux*{.chip data-chip="SYMP neg=true"}, | Le patient n'est pas *fièvreux*{.chip data-chip="SYMP neg=true"}, |
| son père a *du diabète*{.chip data-chip="DIS carrier=PATIENT"}.   | son père a *du diabète*{.chip data-chip="DIS carrier=FATHER"}.    |
| Pas d'évolution du                                                | Pas d'évolution du                                                |
| *cancer*{.chip data-chip="DIS neg=true carrier=PATIENT"}.         | *cancer*{.chip data-chip="DIS carrier=PATIENT"}.                  |
+-------------------------------------------------------------------+-------------------------------------------------------------------+

We can quickly create matching documents in EDS-NLP using the following code snippet:

```python
from edsnlp.data.converters import MarkupToDocConverter

conv = MarkupToDocConverter(preset="md", span_setter="entities")
# Create a document with predicted attributes and a reference document
pred = conv(
    "Le patient n'est pas [fièvreux](SYMP neg=true), "
    "son père a [du diabète](DIS neg=false carrier=PATIENT). "
    "Pas d'évolution du [cancer](DIS neg=true carrier=PATIENT)."
)
ref = conv(
    "Le patient n'est pas [fièvreux](SYMP neg=true), "
    "son père a [du diabète](DIS neg=false carrier=FATHER). "
    "Pas d'évolution du [cancer](DIS neg=false carrier=PATIENT)."
)
```
"""  # noqa: E501

import warnings
from collections import defaultdict
from typing import Any, Dict, Optional

from edsnlp import registry
from edsnlp.metrics import Examples, average_precision, make_examples, prf
from edsnlp.utils.bindings import BINDING_GETTERS, Attributes, AttributesArg
from edsnlp.utils.span_getters import SpanGetterArg, get_spans


def span_attribute_metric(
    examples: Examples,
    span_getter: SpanGetterArg,
    attributes: Attributes = None,
    include_falsy: bool = False,
    default_values: Dict = {},
    micro_key: str = "micro",
    filter_expr: Optional[str] = None,
    **kwargs: Any,
):
    if "qualifiers" in kwargs:
        warnings.warn(
            "The `qualifiers` argument of span_attribute_metric() is "
            "deprecated. Use `attributes` instead.",
            DeprecationWarning,
        )
        assert attributes is None
        attributes = kwargs.pop("qualifiers")
    if attributes is None:
        raise TypeError(
            "span_attribute_metric() missing 1 required argument: 'attributes'"
        )
    if kwargs:
        raise TypeError(
            f"span_attribute_metric() got unexpected keyword arguments: "
            f"{', '.join(kwargs.keys())}"
        )
    examples = make_examples(examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]
    labels = defaultdict(lambda: (set(), set(), dict()))
    labels["micro"] = (set(), set(), dict())
    total_pred_count = 0
    total_gold_count = 0

    if not include_falsy:
        default_values_ = defaultdict(lambda: False)
        default_values_.update(default_values)
        default_values = default_values_
        del default_values_
    for eg_idx, eg in enumerate(examples):
        doc_spans = get_spans(eg.predicted, span_getter)
        for span in doc_spans:
            total_pred_count += 1
            beg, end = span.start, span.end
            for attr, span_filter in attributes.items():
                if not (span_filter is True or span.label_ in span_filter):
                    continue
                getter_key = attr if attr.startswith("_.") else f"_.{attr}"
                value = BINDING_GETTERS[getter_key](span)
                top_val, top_p = max(
                    getattr(span._, "prob", {}).get(attr, {}).items(),
                    key=lambda x: x[1],
                    default=(value, 1.0),
                )
                if (top_val or include_falsy) and default_values[attr] != top_val:
                    labels[attr][2][(eg_idx, beg, end, attr, top_val)] = top_p
                    labels[micro_key][2][(eg_idx, beg, end, attr, top_val)] = top_p
                if (value or include_falsy) and default_values[attr] != value:
                    labels[micro_key][0].add((eg_idx, beg, end, attr, value))
                    labels[attr][0].add((eg_idx, beg, end, attr, value))

        doc_spans = get_spans(eg.reference, span_getter)
        for span in doc_spans:
            total_gold_count += 1
            beg, end = span.start, span.end
            for attr, span_filter in attributes.items():
                if not (span_filter is True or span.label_ in span_filter):
                    continue
                getter_key = attr if attr.startswith("_.") else f"_.{attr}"
                value = BINDING_GETTERS[getter_key](span)
                if (value or include_falsy) and default_values[attr] != value:
                    labels[micro_key][1].add((eg_idx, beg, end, attr, value))
                    labels[attr][1].add((eg_idx, beg, end, attr, value))

    if total_pred_count != total_gold_count:
        raise ValueError(
            f"Number of predicted and gold spans differ: {total_pred_count} != "
            f"{total_gold_count}. Make sure that you are running your span "
            "attribute classification pipe on the gold annotations, and not spans "
            "predicted by another NER pipe in your model."
        )

    return {
        name: {
            **prf(pred, gold),
            "ap": average_precision(pred_with_prob, gold),
        }
        for name, (pred, gold, pred_with_prob) in labels.items()
    }


@registry.metrics.register(
    "eds.span_attribute",
    deprecated=["eds.span_classification_scorer", "eds.span_attribute_scorer"],
)
class SpanAttributeMetric:
    """
    The `eds.span_attribute` metric
    evaluates span‐level attribute classification by comparing predicted and gold
    attribute values on the same set of spans. For each attribute you specify, it
    computes Precision, Recall, F1, number of true positives (tp), number of
    gold instances (support), number of predicted instances (positives), and
    the Average Precision (ap). A micro‐average over all attributes is also
    provided under `micro_key`.

    ```python { .no-check }
    from edsnlp.metrics.span_attribute import SpanAttributeMetric

    metric = SpanAttributeMetric(
        span_getter=conv.span_setter,
        # Evaluated attributes
        attributes={
            "neg": True,  # 'neg' on every entity
            "carrier": ["DIS"],  # 'carrier' only on 'DIS' entities
        },
        # Ignore these default values when counting matches
        default_values={
            "neg": False,
        },
        micro_key="micro",
    )
    ```

    Let's enumerate (span -> attr = value) items in our documents. Only the items with
    matching span boundaries, attribute name, and value are counted as a true positives.
    For instance, with the predicted and reference spans of the example above:

    +--------------------------------------------------+-------------------------------------------------+
    | pred                                             | ref                                             |
    +==================================================+=================================================+
    | *fièvreux → neg = True*{.chip .tp}<br/>          | *fièvreux → neg = True*{.chip .tp}<br/>         |
    | *du diabète → neg = False*{.chip .na}<br/>       | *du diabète → neg = False*{.chip .na}<br/>      |
    | *du diabète → carrier = PATIENT*{.chip .fp}<br/> | *du diabète → carrier = FATHER*{.chip .fn}<br/> |
    | *cancer → neg = True*{.chip .fp}<br/>            | *cancer → neg = False*{.chip .na}<br/>          |
    | *cancer → carrier = PATIENT*{.chip .tp}          | *cancer → carrier = PATIENT*{.chip .tp}         |
    +--------------------------------------------------+-------------------------------------------------+

    !!! note "Default values"

        Note that there we don't count "neg=False" items, shown in grey in the table. In EDS-NLP,
        this is done by setting `defaults_values={"neg": False}` when creating the metric. This
        is quite common in classification tasks, where one of the values is both the most common
        and the "default" (hence the name of the parameter). Counting these values would likely
        skew the micro-average metrics towards the default value.

    Precision, Recall and F1 (micro-average and per‐label) are computed as follows:

    - Precision: `p = |matched items of pred| / |pred|`
    - Recall: `r = |matched items of ref| / |ref|`
    - F1: `f = 2 / (1/p + 1/f)`

    This yields the following metrics:

    ```python
    metric([ref], [pred])
    # Out: {
    #   'micro': {'f': 0.57, 'p': 0.5, 'r': 0.67, 'tp': 2, 'support': 3, 'positives': 4, 'ap': 0.17},
    #   'neg': {'f': 0.67, 'p': 0.5, 'r': 1, 'tp': 1, 'support': 1, 'positives': 2, 'ap': 0.0},
    #   'carrier': {'f': 0.5, 'p': 0.5, 'r': 0.5, 'tp': 1, 'support': 2, 'positives': 2, 'ap': 0.25},
    # }
    ```

    Parameters
    ----------
    span_getter : SpanGetterArg
        The span getter to extract spans from each `Doc`.
    attributes : Mapping[str, Union[bool, Sequence[str]]]
        Map each attribute name to `True` (evaluate on all spans) or a sequence of
        labels restricting which spans to test.
    default_values : Dict[str, Any]
        Attribute values to omit from micro‐average counts (e.g., common negative or
        default labels).
    include_falsy : bool
        If `False`, ignore falsy values (e.g., `False`, `None`, `''`) in predictions
        or gold when computing metrics; if `True`, count them.
    micro_key : str
        Key under which to store the micro‐averaged results across all attributes.
    filter_expr : Optional[str]
        A Python expression (using `doc`) to filter which examples are scored.

    Returns
    -------
    Dict[str, Dict[str, float]]
        A dictionary mapping each attribute name (and the `micro_key`) to its metrics:

        - `label` or micro_key :

            - `p` : precision
            - `r` : recall
            - `f` : F1 score
            - `tp` : true positive count
            - `support` : number of gold instances
            - `positives` : number of predicted instances
            - `ap` : [average precision](https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision)
    """  # noqa: E501

    attributes: Attributes

    def __init__(
        self,
        span_getter: SpanGetterArg,
        attributes: AttributesArg = None,
        qualifiers: AttributesArg = None,
        default_values: Dict = {},
        include_falsy: bool = False,
        micro_key: str = "micro",
        filter_expr: Optional[str] = None,
    ):
        if qualifiers is not None:
            warnings.warn(
                "The `qualifiers` argument is deprecated. Use `attributes` instead.",
                DeprecationWarning,
            )
        self.span_getter = span_getter
        self.attributes = attributes or qualifiers
        self.default_values = default_values
        self.include_falsy = include_falsy
        self.micro_key = micro_key
        self.filter_expr = filter_expr

    __init__.__doc__ = span_attribute_metric.__doc__

    def __call__(self, *examples: Any):
        """
        Compute the span attribute metrics for the given examples.

        Parameters
        ----------
        examples : Examples
            The examples to score, either a tuple of (golds, preds) or a list of
            spacy.training.Example objects

        Returns
        -------
        Dict[str, Dict[str, float]]
            The scores for the attributes
        """
        return span_attribute_metric(
            examples,
            span_getter=self.span_getter,
            attributes=self.attributes,
            default_values=self.default_values,
            include_falsy=self.include_falsy,
            micro_key=self.micro_key,
            filter_expr=self.filter_expr,
        )


# For backward compatibility
span_classification_scorer = span_attribute_scorer = span_attribute_metric
create_span_attributes_scorer = SpanAttributeScorer = SpanAttributeMetric

__all__ = [
    "span_attribute_metric",
    "SpanAttributeMetric",
]



from __future__ import annotations

import sys
from contextlib import nullcontext

from tqdm import tqdm

from edsnlp.core.stream import Stream, StreamSentinel

doc_size_fns = {
    "words": len,
}


def execute_simple_backend(stream: Stream):
    """
    This is the default execution mode which batches the documents and processes each
    batch on the current process in a sequential manner.
    """
    try:
        torch = sys.modules["torch"]
        no_grad_ctx = torch.no_grad()
        device = next(
            (p.device for pipe in stream.torch_components() for p in pipe.parameters()),
            torch.device("cpu"),
        )
        device_type = getattr(device, "type", device).split(":")[0]
        autocast = stream.autocast
        autocast_ctx = nullcontext()
        try:
            if autocast:
                autocast_ctx = torch.autocast(
                    device_type=device_type,
                    dtype=autocast if autocast is not True else None,
                )
        except RuntimeError:  # pragma: no cover
            pass

        inference_mode_ctx = (
            torch.inference_mode()
            if hasattr(torch, "inference_mode")
            else nullcontext()
        )
    except (KeyError, AttributeError):  # pragma: no cover
        no_grad_ctx = autocast_ctx = inference_mode_ctx = nullcontext()
    reader = stream.reader
    writer = stream.writer
    show_progress = stream.show_progress
    stages = stream._make_stages(split_torch_pipes=True)

    def make_torch_pipe(torch_pipe, disable_after):
        def wrapped(batches):
            for batch in batches:
                with autocast_ctx, inference_mode_ctx, no_grad_ctx:
                    batch_id = hash(tuple(id(x) for x in batch))
                    torch_pipe.enable_cache(batch_id)
                    batch = torch_pipe.batch_process(batch)
                    if disable_after:
                        torch_pipe.disable_cache(batch_id)
                yield batch

        return wrapped

    def process():
        bar = tqdm(smoothing=0.1, mininterval=5.0, disable=not show_progress)

        with bar, stream.eval():
            items = reader.read_records()
            items = (
                task
                for item in items
                for task in (
                    (item,)
                    if isinstance(item, StreamSentinel)
                    else reader.extract_task(item)
                )
            )

            for stage_idx, stage in enumerate(stages):
                for op in stage.cpu_ops:
                    items = op(items)

                if stage.gpu_op is not None:
                    pipe = make_torch_pipe(stage.gpu_op, stage_idx == len(stages) - 2)
                    items = pipe(items)

            if writer is not None:
                items = (
                    item
                    if isinstance(item, StreamSentinel)
                    else writer.handle_record(item)
                    for item in items
                )

            if getattr(writer, "batch_fn", None) is not None:
                items = writer.batch_fn(items, writer.batch_size, sentinel_mode="drop")
                # get the 1st element (2nd is the count)
                for b in items:
                    if not isinstance(b, StreamSentinel):
                        item, count = writer.handle_batch(b)
                        bar.update(count)
                    yield item
            else:
                for item in items:
                    if not isinstance(item, StreamSentinel):
                        bar.update(1)
                        yield item

    items = process()

    if writer is not None:
        items = writer.consolidate(items)

    return items



import pytest
from spacy.tokens import Span

import edsnlp
from edsnlp.scorers.ner import NerExactScorer, NerOverlapScorer, NerTokenScorer
from edsnlp.scorers.span_attributes import SpanAttributeScorer


@pytest.fixture(scope="session")
def gold_and_pred():
    nlp = edsnlp.blank("eds")

    gold_doc1 = nlp.make_doc("Le patient a le covid 19.")
    gold_doc1.ents = [
        Span(gold_doc1, 4, 6, label="covid"),  # le covid
    ]
    pred_doc1 = nlp.make_doc("Le patient a le covid 19.")
    pred_doc1.ents = [
        Span(pred_doc1, 4, 6, label="covid"),  # le covid
    ]
    gold_doc2 = nlp.make_doc(
        "Corona: positif. Le cvid est une maladie très très grave."
    )
    gold_doc2.ents = [
        Span(gold_doc2, 0, 1, label="covid"),  # Corona
        Span(gold_doc2, 5, 6, label="covid"),  # cvid
        Span(gold_doc2, 8, 12, label="disease"),  # maladie très très grave
    ]

    pred_doc2 = nlp.make_doc(
        "Corona: positif. Le cvid est une maladie très très grave."
    )
    pred_doc2.ents = [
        Span(pred_doc2, 0, 2, label="covid"),  # Corona:
        Span(pred_doc2, 8, 9, label="disease"),  # maladie
    ]

    return [gold_doc1, gold_doc2], [pred_doc1, pred_doc2]


def test_exact_ner_scorer(gold_and_pred):
    scorer = NerExactScorer("ents", filter_expr="'vid' in doc.text")
    ner_exact_score = scorer(*gold_and_pred)
    assert ner_exact_score["micro"] == {
        "f": 0.2857142857142857,
        "p": 0.3333333333333333,
        "positives": 3,
        "r": 0.25,
        "support": 4,
        "tp": 1,
    }


def test_token_ner_scorer(gold_and_pred):
    scorer = NerTokenScorer("ents", filter_expr="'vid' in doc.text")
    ner_exact_score = scorer(*gold_and_pred)
    assert ner_exact_score["micro"] == {
        "f": 0.6153846153846154,
        "p": 0.8,
        "positives": 5,
        "r": 0.5,
        "support": 8,
        "tp": 4,
    }


def test_overlap_ner_scorer_any(gold_and_pred):
    scorer = NerOverlapScorer(
        "ents", threshold=0.00001, filter_expr="'vid' in doc.text"
    )
    # pred entities: [le covid, Corona:, maladie] => 3
    # gold entities: [le covid, Corona, cvid, maladie très très grave] => 4
    # tp: [le covid, Corona, maladie] => 3
    ner_exact_score = scorer(*gold_and_pred)
    assert ner_exact_score["micro"] == {
        "f": 0.8571428571428572,
        "p": 1,
        "positives": 3,
        "r": 0.75,
        "support": 4,
        "tp": 3,
    }


def test_overlap_ner_scorer_half(gold_and_pred):
    scorer = NerOverlapScorer(
        "ents",
        threshold=0.5,
        filter_expr="'vid' in doc.text",
    )
    # pred entities: [le covid, Corona:, maladie] => 3
    # gold entities: [le covid, Corona, cvid, maladie très très grave] => 4
    # tp: [le covid, Corona] => 2
    ner_exact_score = scorer(*gold_and_pred)
    assert ner_exact_score["micro"] == {
        "f": 0.5714285714285714,
        "p": 0.6666666666666666,
        "positives": 3,
        "r": 0.5,
        "support": 4,
        "tp": 2,
    }


def test_overlap_ner_scorer_full(gold_and_pred):
    scorer = NerOverlapScorer(
        "ents",
        threshold=1.0,
        filter_expr="'vid' in doc.text",
    )
    ner_exact_score = scorer(*gold_and_pred)
    # pred entities: [le covid, Corona:, maladie] => 3
    # gold entities: [le covid, Corona, cvid, maladie très très grave] => 4
    # tp: [le covid] => 2
    assert ner_exact_score["micro"] == {
        "f": 0.2857142857142857,
        "p": 0.3333333333333333,
        "positives": 3,
        "r": 0.25,
        "support": 4,
        "tp": 1,
    }


def test_span_attributes_scorer():
    if not Span.has_extension("negation"):
        Span.set_extension("negation", default=False)
    pred = edsnlp.blank("eds")("Le patient n'a pas le covid 19.")
    gold = edsnlp.blank("eds")("Le patient n'a pas le covid 19.")
    scorer = SpanAttributeScorer(
        "entities",
        "negation",
        default_values={"negation": False},
        filter_expr="'vid' in doc.text",
    )
    pred.spans["entities"] = [
        pred[1:2],
        pred[3:4],
    ]
    pred.spans["entities"][0]._.negation = True
    pred.spans["entities"][1]._.negation = True
    gold.spans["entities"] = [
        gold[1:2],
        gold[3:4],
    ]
    gold.spans["entities"][0]._.negation = False
    gold.spans["entities"][1]._.negation = True
    result = scorer([gold], [pred])
    assert result["micro"] == {
        "ap": 0.5,
        "p": 0.5,
        "r": 1,
        "f": 2 / 3,
        "support": 1,
        "positives": 2,
        "tp": 1,
    }



"""
We provide several metrics to evaluate the performance of Named Entity Recognition (NER) components.
Let's look at an example and see how they differ. We'll use the following two documents: a reference
document (ref) and a document with predicted entities (pred).

+-------------------------------------------------------------+------------------------------------------+
| pred                                                        | ref                                      |
+=============================================================+==========================================+
| *La*{.chip data-chip=PER} *patiente*{.chip data-chip=PER} a | La *patiente*{.chip data-chip=PER} a     |
| une *fièvre aigüe*{.chip data-chip=DIS}                     | *une fièvre*{.chip data-chip=DIS} aigüe. |
+-------------------------------------------------------------+------------------------------------------+

Let's create matching documents in EDS-NLP using the following code snippet:

```python
from edsnlp.data.converters import MarkupToDocConverter

conv = MarkupToDocConverter(preset="md", span_setter="entities")

pred = conv("[La](PER) [patiente](PER) a une [fièvre aiguë](DIS).")
ref = conv("La [patiente](PER) a [une fièvre](DIS) aiguë.")
```
"""  # noqa: E501

import abc
from collections import defaultdict
from typing import Any, Dict, Optional

from edsnlp import registry
from edsnlp.metrics import Examples, make_examples, prf
from edsnlp.utils.span_getters import SpanGetter, SpanGetterArg, get_spans


def ner_exact_metric(
    examples: Examples,
    span_getter: SpanGetter,
    micro_key: str = "micro",
    filter_expr: Optional[str] = None,
) -> Dict[str, Any]:
    examples = make_examples(examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]
    labels = defaultdict(lambda: (set(), set()))
    labels["micro"] = (set(), set())
    for eg_idx, eg in enumerate(examples):
        for span in (
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        ):
            labels[span.label_][0].add((eg_idx, span.start, span.end, span.label_))
            labels[micro_key][0].add((eg_idx, span.start, span.end, span.label_))

        for span in (
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        ):
            labels[span.label_][1].add((eg_idx, span.start, span.end, span.label_))
            labels[micro_key][1].add((eg_idx, span.start, span.end, span.label_))

    return {name: prf(pred, gold) for name, (pred, gold) in labels.items()}


def ner_token_metric(
    examples: Examples,
    span_getter: SpanGetter,
    micro_key: str = "micro",
    filter_expr: Optional[str] = None,
) -> Dict[str, Any]:
    examples = make_examples(examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]
    # label -> pred, gold
    labels = defaultdict(lambda: (set(), set()))
    labels["micro"] = (set(), set())
    for eg_idx, eg in enumerate(examples):
        for span in (
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        ):
            for i in range(span.start, span.end):
                labels[span.label_][0].add((eg_idx, i, span.label_))
                labels[micro_key][0].add((eg_idx, i, span.label_))

        for span in (
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        ):
            for i in range(span.start, span.end):
                labels[span.label_][1].add((eg_idx, i, span.label_))
                labels[micro_key][1].add((eg_idx, i, span.label_))

    return {name: prf(pred, gold) for name, (pred, gold) in labels.items()}


def dice(span1, span2):
    """
    Compute the Dice coefficient between two spans
    """
    intersection = max(0, min(span1.end, span2.end) - max(span1.start, span2.start))
    return 2 * intersection / (span1.end - span1.start + span2.end - span2.start)


def ner_overlap_metric(
    examples: Examples,
    span_getter: SpanGetter,
    micro_key: str = "micro",
    filter_expr: Optional[str] = None,
    threshold: float = 0.5,
) -> Dict[str, Any]:
    examples = make_examples(*examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]
    # label -> pred, gold, matched_pred, matched_gold
    counters = defaultdict(lambda: [0, 0, 0, 0])
    counters["micro"] = [0, 0, 0, 0]
    for eg_idx, eg in enumerate(examples):
        pred_spans = set(
            span_getter(eg.predicted)
            if callable(span_getter)
            else get_spans(eg.predicted, span_getter)
        )
        gold_spans = set(
            span_getter(eg.reference)
            if callable(span_getter)
            else get_spans(eg.reference, span_getter)
        )
        for gold_span in gold_spans:
            counters[gold_span.label_][1] += 1
            counters[micro_key][1] += 1
        for pred_span in pred_spans:
            counters[pred_span.label_][0] += 1
            counters[micro_key][0] += 1

        for pred_span in pred_spans:
            overlaps = [
                (gold_span, dice(pred_span, gold_span))
                for gold_span in gold_spans
                if gold_span.label == pred_span.label
            ]
            matching_gold_span, overlap = max(
                overlaps, key=lambda x: x[1], default=(None, 0)
            )
            if matching_gold_span is not None and overlap >= threshold:
                counters[pred_span.label_][2] += 1
                counters[micro_key][2] += 1

        for gold_span in gold_spans:
            overlaps = [
                (pred_span, dice(gold_span, pred_span)) for pred_span in pred_spans
            ]
            matching_pred_span, overlap = max(
                overlaps, key=lambda x: x[1], default=(None, 0)
            )
            if (
                matching_pred_span is not None
                and overlap >= threshold
                and gold_span.label == matching_pred_span.label
            ):
                counters[gold_span.label_][3] += 1
                counters[micro_key][3] += 1

    results = {}
    for name, (num_pred, num_gold, true_pred, true_gold) in counters.items():
        p = (true_pred / num_pred) if num_pred else 1.0
        r = (true_gold / num_gold) if num_gold else 1.0
        f = (
            2 / (num_pred / true_pred + num_gold / true_gold)
            if true_pred and true_gold
            else 0.0
        )
        results[name] = {
            "f": f,
            "p": p,
            "r": r,
            "tp": true_pred,
            "support": num_gold,  # num gold
            "positives": num_pred,  # num predicted
        }
    return results


class NerMetric(abc.ABC):
    span_getter: SpanGetter

    def __call__(self, *examples) -> Dict[str, Any]:
        raise NotImplementedError()


@registry.metrics.register(
    "eds.ner_exact",
    deprecated=["eds.ner_exact_metric"],
)
class NerExactMetric(NerMetric):
    r"""
    The `eds.ner_exact` metric
    scores the extracted entities (that may be overlapping or nested)
    by looking in the spans returned by a given SpanGetter object and
    comparing predicted spans to gold spans for **exact** boundary and label matches.

    Let's view these elements as collections of (span → label) and count how
    many of the predicted spans match the gold spans exactly (and vice versa):

    +----------------------------------------------+--------------------------------------------+
    | pred                                         | ref                                        |
    +==============================================+============================================+
    | *La*{.chip .fp data-chip=PER}<br/>           | *patiente*{.chip .tp data-chip=PER}<br/>   |
    | *patiente*{.chip .tp data-chip=PER}<br/>     | *une fièvre*{.chip .fp data-chip=DIS}<br/> |
    | *fièvre aiguë*{.chip .fp data-chip=DIS}<br/> |                                            |
    +----------------------------------------------+--------------------------------------------+

    Precision, Recall and F1 (micro-average and per‐label) are computed as follows:

    - Precision: `p = |matched items of pred| / |pred|`
    - Recall: `r = |matched items of ref| / |ref|`
    - F1: `f = 2 / (1/p + 1/f)`

    Examples
    --------

    ```python
    from edsnlp.metrics.ner import NerExactMetric

    metric = NerExactMetric(span_getter=conv.span_setter, micro_key="micro")
    metric([ref], [pred])
    # Out: {
    #   'micro': {'f': 0.4, 'p': 0.33, 'r': 0.5, 'tp': 1, 'support': 2, 'positives': 3},
    #   'PER': {'f': 0.67, 'p': 0.5, 'r': 1, 'tp': 1, 'support': 1, 'positives': 2},
    #   'DIS': {'f': 0.0, 'p': 0.0, 'r': 0.0, 'tp': 0, 'support': 1, 'positives': 1},
    # }
    ```

    Parameters
    ----------
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types
    filter_expr: str
        The filter expression to use to filter the documents. Evaluated with `doc` as the variable.
    """  # noqa: E501

    def __init__(
        self,
        span_getter: SpanGetterArg,
        micro_key: str = "micro",
        filter_expr: Optional[str] = None,
    ):
        self.span_getter = span_getter
        self.micro_key = micro_key
        self.filter_expr = filter_expr

    __init__.__doc__ = ner_exact_metric.__doc__

    def __call__(self, *examples):
        return ner_exact_metric(
            examples,
            span_getter=self.span_getter,
            micro_key=self.micro_key,
            filter_expr=self.filter_expr,
        )


@registry.metrics.register(
    "eds.ner_token",
    deprecated=["eds.ner_token_metric"],
)
class NerTokenMetric(NerMetric):
    r"""
    The `eds.ner_token` metric
    scores the extracted entities that may be overlapping or nested by looking in
    `doc.ents`, and `doc.spans`, and comparing the predicted and gold entities at the
    **token** level.

    Assuming we use the `eds` (or `fr` or `en`) tokenizer, in the above example, there
    are 3 annotated tokens in the reference, and 4 annotated tokens in the prediction.
    Let's view these elements as sets of (token, label) and count how many of the
    predicted tokens match the gold tokens exactly (and vice versa):

    +------------------------------------------+------------------------------------------+
    | pred                                     | ref                                      |
    +==========================================+==========================================+
    | *La*{.chip .fp data-chip=PER}<br/>       | *patiente*{.chip .tp data-chip=PER}<br/> |
    | *patiente*{.chip .tp data-chip=PER}<br/> | *une*{.chip .fp data-chip=DIS}<br/>      |
    | *fièvre*{.chip .tp data-chip=DIS}<br/>   | *fièvre*{.chip .tp data-chip=DIS}        |
    | *aiguë*{.chip .fp data-chip=DIS}         |                                          |
    +------------------------------------------+------------------------------------------+

    Precision, Recall and F1 (micro-average and per‐label) are computed as follows:

    - Precision: `p = |matched items of pred| / |pred|`
    - Recall: `r = |matched items of ref| / |ref|`
    - F1: `f = 2 / (1/p + 1/f)`

    Examples
    --------

    ```python
    from edsnlp.metrics.ner import NerTokenMetric

    metric = NerTokenMetric(span_getter=conv.span_setter, micro_key="micro")
    metric([ref], [pred])
    # Out: {
    #   'micro': {'f': 0.57, 'p': 0.5, 'r': 0.67, 'tp': 2, 'support': 3, 'positives': 4},
    #   'PER': {'f': 0.67, 'p': 0.5, 'r': 1, 'tp': 1, 'support': 1, 'positives': 2},
    #   'DIS': {'f': 0.5, 'p': 0.5, 'r': 0.5, 'tp': 1, 'support': 2, 'positives': 2}
    # }
    ```

    Parameters
    ----------
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types
    filter_expr: str
        The filter expression to use to filter the documents. Will be evaluated
        with `doc` as the variable name, so you can use `doc.ents`, `doc.spans`, etc.
    """  # noqa: E501

    def __init__(
        self,
        span_getter: SpanGetterArg,
        micro_key: str = "micro",
        filter_expr: Optional[str] = None,
    ):
        self.span_getter = span_getter
        self.micro_key = micro_key
        self.filter_expr = filter_expr

    __init__.__doc__ = ner_token_metric.__doc__

    def __call__(self, *examples):
        return ner_token_metric(
            examples,
            span_getter=self.span_getter,
            micro_key=self.micro_key,
            filter_expr=self.filter_expr,
        )


@registry.metrics.register(
    "eds.ner_overlap",
    deprecated=["eds.ner_overlap_metric"],
)
class NerOverlapMetric(NerMetric):
    r"""
    The `eds.ner_overlap` metric
    scores the extracted entities that may be overlapping or nested
    by looking in the spans returned by a given SpanGetter object and
    counting a prediction as correct if it overlaps by at least the given
    Dice‐coefficient threshold with a gold span of the same label.

    This metric is useful for evaluating NER systems where the exact boundaries
    do not matter too much, but the presence of the entity at the same spot is important.
    For instance, you may not want to penalize a system that forgets determiners if
    the rest of the entity is correctly identified.

    Let's view these elements as sets of (span → label) and count how many of the
    predicted spans match the gold spans by at least the given Dice coefficient
    (and vice versa):

    +---------------------------------------------+------------------------------------------+
    | pred                                        | ref                                      |
    +=============================================+==========================================+
    | *La*{.chip .fp data-chip=PER}<br/>          | *patiente*{.chip .tp data-chip=PER}<br/> |
    | *patiente*{.chip .tp data-chip=PER}<br/>    | *une fièvre*{.chip .tp data-chip=DIS}    |
    | *fièvre aiguë*{.chip .tp data-chip=DIS}<br/>|                                          |
    +---------------------------------------------+------------------------------------------+

    Precision, Recall and F1 (micro-average and per‐label) are computed as follows:

    - Precision: `p = |matched items of pred| / |pred|`
    - Recall: `r = |matched items of ref| / |ref|`
    - F1: `f = 2 / (1/p + 1/f)`

    !!! note "Overlap threshold"

        The threshold is the minimum Dice coefficient to consider two spans as overlapping. Setting
        it to 1.0 will yield the same results as the `eds.ner_exact` metric, while setting it to a
        near-zero value (e.g., like 1e-14) will match any two spans that share at least one token.

    Examples
    --------

    ```python
    from edsnlp.metrics.ner import NerOverlapMetric

    metric = NerOverlapMetric(
        span_getter=conv.span_setter, micro_key="micro", threshold=0.5
    )
    metric([ref], [pred])
    # Out: {
    #   'micro': {'f': 0.8, 'p': 0.67, 'r': 1.0, 'tp': 2, 'support': 2, 'positives': 3},
    #   'PER': {'f': 0.67, 'p': 0.5, 'r': 1.0, 'tp': 1, 'support': 1, 'positives': 2},
    #   'DIS': {'f': 1.0, 'p': 1.0, 'r': 1.0, 'tp': 1, 'support': 1, 'positives': 1}
    # }
    ```

    Parameters
    ----------
    span_getter: SpanGetter
        The span getter to use to extract the spans from the document
    micro_key: str
        The key to use to store the micro-averaged results for spans of all types
    filter_expr: str
        The filter expression to use to filter the documents
    threshold: float
        The threshold on the Dice coefficient to consider two spans as overlapping
    """  # noqa: E501

    def __init__(
        self,
        span_getter: SpanGetterArg,
        micro_key: str = "micro",
        filter_expr: Optional[str] = None,
        threshold: float = 0.5,
    ):
        self.span_getter = span_getter
        self.micro_key = micro_key
        self.filter_expr = filter_expr
        self.threshold = threshold

    __init__.__doc__ = ner_overlap_metric.__doc__

    def __call__(self, *examples):
        return ner_overlap_metric(
            examples,
            span_getter=self.span_getter,
            micro_key=self.micro_key,
            filter_expr=self.filter_expr,
            threshold=self.threshold,
        )


# For backward compatibility
ner_exact_scorer = ner_exact_metric
ner_token_scorer = ner_token_metric
ner_overlap_scorer = ner_overlap_metric
create_ner_exact_scorer = NerExactScorer = NerExactMetric
create_ner_token_scorer = NerTokenScorer = NerTokenMetric
create_ner_overlap_scorer = NerOverlapScorer = NerOverlapMetric

__all__ = [
    "NerMetric",
    "NerExactMetric",
    "NerTokenMetric",
    "NerOverlapMetric",
    "ner_exact_metric",
    "ner_token_metric",
    "ner_overlap_metric",
]



from edsnlp import registry

from .biaffine_dep_parser import TrainableBiaffineDependencyParser

create_component = registry.factory.register(
    "eds.biaffine_dep_parser",
    assigns=["token.head", "token.dep"],
)(TrainableBiaffineDependencyParser)
